############################################################################
#  imagecomp.py
#
#  Copyright 2017 RSGISLib.
#
#  RSGISLib: 'The remote sensing and GIS Software Library'
#
#  RSGISLib is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  RSGISLib is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with RSGISLib.  If not, see <http://www.gnu.org/licenses/>.
#
#
# Purpose:  Provide a set of utilities for creating image composites
#
# Author: Pete Bunting
# Email: petebunting@mac.com
# Date: 04/10/2017
# Version: 1.0
#
# History:
# Version 1.0 - Created.
#
############################################################################

import rsgislib
import rsgislib.imagecalc
import rsgislib.imagecalc.calcindices
import rsgislib.imageutils
import rsgislib.rastergis

import glob
import os
import os.path
import numpy
import shutil

import osgeo.gdal
import rios.rat


def createMaxNDVIComposite(inImgsPattern, rBand, nBand, outRefImg, outCompImg, tmpPath='./tmp', gdalformat='KEA', dataType=None, calcStats=True):
    """
Create an image composite from multiple input images where the pixel brought through into the composite is the one with
the maximum NDVI.

:param inImgsPattern: is a pattern (ready for glob.glob(inImgsPattern)) so needs an '*' within the pattern to find the input image files.
:param rBand: is the image band within the input images (same for all) for the red band - note band indexing starts at 1.
:param nBand: is the image band within the input images (same for all) for the nir band - note band indexing starts at 1.
:param outRefImg: is the output reference image which details which input image is forming the output image pixel value (Note. this file will always be a KEA file as RAT is used).
:param outCompImg: is the output composite image for which gdalformat and dataType define the format and data type.
:param tmpPath: is a temp path for intemediate files, if this path doesn't exist is will be created and deleted at runtime.
:param gdalformat: is the output file format of the outCompImg, any GDAL compatable format is OK (Defaut is KEA).
:param dataType: is the data type of the output image (outCompImg). If None is provided then the data type of the first input image will be used (Default None).
:param calcStats: calculate image statics and pyramids (Default=True)

    """
    rsgisUtils = rsgislib.RSGISPyUtils()
    uidStr = rsgisUtils.uidGenerator()
    
    # Get List of input images:
    inImages = glob.glob(inImgsPattern)
    
    if len(inImages) > 1:
        tmpPresent = True
        if not os.path.exists(tmpPath):
            os.makedirs(tmpPath)
            tmpPresent = False 
        
        refLayersPath = os.path.join(tmpPath, 'RefLyrs_'+uidStr)
        
        refImgTmpPresent = True
        if not os.path.exists(refLayersPath):
            os.makedirs(refLayersPath)
            refImgTmpPresent = False
        
        if dataType is None:
            rsgisUtils.getRSGISLibDataTypeFromImg(inImages[0])
        
        numInLyrs = len(inImages)
        
        numpy.random.seed(5)
        red = numpy.random.randint(1, 255, numInLyrs+1, int)
        red[0] = 0
        numpy.random.seed(2)
        green = numpy.random.randint(1, 255, numInLyrs+1, int)
        green[0] = 0
        numpy.random.seed(9)
        blue = numpy.random.randint(1, 255, numInLyrs+1, int)
        blue[0] = 0
        alpha = numpy.zeros_like(blue)
        alpha[...] = 255
        imgLyrs = numpy.empty(numInLyrs+1, dtype=numpy.dtype('a255'))
            
        # Generate Comp Ref layers:
        refLyrsLst = []
        idx = 1
        for img in inImages:
            print('In Image ('+str(idx) + '):\t' + img)
            imgLyrs[idx] = os.path.basename(img)
            baseImgName = os.path.splitext(os.path.basename(img))[0]
            refLyrImg = os.path.join(refLayersPath, baseImgName+'_ndvi.kea')
            rsgislib.imagecalc.calcindices.calcNDVI(img, rBand, nBand, refLyrImg, False)
            refLyrsLst.append(refLyrImg)
            idx = idx + 1
        imgLyrs[0] = ""
    
        # Create REF Image
        rsgislib.imagecalc.getImgIdxForStat(refLyrsLst, outRefImg, 'KEA', -999, rsgislib.SUMTYPE_MAX)
        if calcStats:
            # Pop Ref Image with stats
            rsgislib.rastergis.populateStats(outRefImg, True, True, True)
            
            # Open the clumps dataset as a gdal dataset
            ratDataset = osgeo.gdal.Open(outRefImg, osgeo.gdal.GA_Update)
            
            # Write colours to RAT
            rios.rat.writeColumn(ratDataset, "Red", red)
            rios.rat.writeColumn(ratDataset, "Green", green)
            rios.rat.writeColumn(ratDataset, "Blue", blue)
            rios.rat.writeColumn(ratDataset, "Alpha", alpha)
            rios.rat.writeColumn(ratDataset, "Image", imgLyrs)
            
            ratDataset = None
        
        # Create Composite Image
        rsgislib.imageutils.createRefImgCompositeImg(inImages, outCompImg, outRefImg, gdalformat, dataType, 0.0)
        
        if calcStats:
            # Calc Stats
            rsgislib.imageutils.popImageStats(outCompImg, usenodataval=True, nodataval=0, calcpyramids=True)
    
        if not refImgTmpPresent:
            shutil.rmtree(refLayersPath, ignore_errors=True)
    
        if not tmpPresent:
            shutil.rmtree(tmpPath, ignore_errors=True)
    elif len(inImages) == 1:
        print("Only 1 Input Image, Just Copying File to output")
        shutil.copy(inImages[0], outCompImg)
    else:
        raise rsgislib.RSGISPyException("There were no input images for " + inImgsPattern)


def createMaxNDVINDWICompositeLandsat(inImgsPattern, outRefImg, outCompImg, outMskImg, tmpPath='./tmp', gdalformat='KEA', dataType=None, calcStats=True, use_mode=True):
    """
Create an image composite from multiple input images where the pixel brought through into the composite is the one with
the maximum NDVI over land and NDWI over water. A mask of land and water regions is also produced. Landsat 8 images are 
identified as image file names are expected to start 'LS8'. If composite includes a mix of LS8 and LS7 images then the 
LS8 images are submitted to match the images bands of LS7 (i.e., coastal band removed).

:param inImgsPattern: is a pattern (ready for glob.glob(inImgsPattern)) so needs an '*' within the pattern to find the input image files.
:param outRefImg: is the output reference image which details which input image is forming the output image pixel value (Note. this file will always be a KEA file as RAT is used).
:param outCompImg: is the output composite image for which gdalformat and dataType define the format and data type.
:param outMskImg: is the output mask image for regions of water and land where ndvi vs ndwi are used (0=nodata, 1=land, 2=water)
:param tmpPath: is a temp path for intemediate files, if this path doesn't exist is will be created and deleted at runtime.
:param gdalformat: is the output file format of the outCompImg, any GDAL compatable format is OK (Defaut is KEA).
:param dataType: is the data type of the output image (outCompImg). If None is provided then the data type of the first input image will be used (Default None).
:param calcStats: calculate image statics and pyramids (Default=True)
:param use_mode: True: the land/water masks are combined using the mode and False: the land water masks are combined using median.

    """
    rsgisUtils = rsgislib.RSGISPyUtils()
    uidStr = rsgisUtils.uidGenerator()
    
    # Get List of input images:
    inImages = glob.glob(inImgsPattern)
    
    if len(inImages) > 1:
        tmpPresent = True
        if not os.path.exists(tmpPath):
            os.makedirs(tmpPath)
            tmpPresent = False 
        
        refLayersPath = os.path.join(tmpPath, 'RefLyrs_'+uidStr)
        
        refImgTmpPresent = True
        if not os.path.exists(refLayersPath):
            os.makedirs(refLayersPath)
            refImgTmpPresent = False
        
        if dataType is None:
            rsgisUtils.getRSGISLibDataTypeFromImg(inImages[0])
        
        numInLyrs = len(inImages)
        
        numpy.random.seed(5)
        red = numpy.random.randint(1, 255, numInLyrs+1, int)
        red[0] = 0
        numpy.random.seed(2)
        green = numpy.random.randint(1, 255, numInLyrs+1, int)
        green[0] = 0
        numpy.random.seed(9)
        blue = numpy.random.randint(1, 255, numInLyrs+1, int)
        blue[0] = 0
        alpha = numpy.zeros_like(blue)
        alpha[...] = 255
        imgLyrs = numpy.empty(numInLyrs+1, dtype=numpy.dtype('a255'))
            
        # Generate Comp Ref layers:
        landWaterMskBandDefns = []
        mskImgs = []
        idx = 1
        onlyLS8 = True
        incLS8 = False
        for img in inImages:
            print('In Image ('+str(idx) + '):\t' + img)
            imgLyrs[idx] = os.path.basename(img)
            baseImgName = os.path.splitext(os.path.basename(img))[0]
            refLyrNDVIImg = os.path.join(refLayersPath, baseImgName+'_ndvi.kea')
            refLyrNDWIImg = os.path.join(refLayersPath, baseImgName+'_ndwi.kea')
            if 'LS8' in img:
                rBand = 4
                nBand = 5
                sBand = 6
                incLS8 = True
            else:
                rBand = 3
                nBand = 4
                sBand = 5
                onlyLS8 = False
            
            rsgislib.imagecalc.calcindices.calcNDVI(img, rBand, nBand, refLyrNDVIImg, False)
            rsgislib.imagecalc.calcindices.calcNDWI(img, nBand, sBand, refLyrNDWIImg, False)
            
            refLyrMskImg = os.path.join(refLayersPath, baseImgName+'_waterLandMsk.kea')
            bandDefns = []
            bandDefns.append(rsgislib.imagecalc.BandDefn('ndvi', refLyrNDVIImg, 1))
            bandDefns.append(rsgislib.imagecalc.BandDefn('ndwi', refLyrNDWIImg, 1))
            rsgislib.imagecalc.bandMath(refLyrMskImg, 'ndvi<-1?0:ndvi>0.3?1:ndwi>0.01?2:1', 'KEA', rsgislib.TYPE_8UINT, bandDefns)
            mskImgs.append(refLyrMskImg)
            idx = idx + 1
        imgLyrs[0] = ""        
        
        refLyrMskStackImg = os.path.join(refLayersPath, uidStr+'_waterLandMskStack.kea')
        rsgislib.imageutils.stackImageBands(mskImgs, None, refLyrMskStackImg, -1, -1, 'KEA', rsgislib.TYPE_8UINT)
        if use_mode:
            rsgislib.imagecalc.imagePixelColumnSummary(refLyrMskStackImg, outMskImg,
                                                       rsgislib.imagecalc.StatsSummary(calcMode=True), 'KEA',
                                                       rsgislib.TYPE_8UINT, 0, True)
        else:
            rsgislib.imagecalc.imagePixelColumnSummary(refLyrMskStackImg, outMskImg,
                                                       rsgislib.imagecalc.StatsSummary(calcMedian=True), 'KEA',
                                                       rsgislib.TYPE_8UINT, 0, True)
        rsgislib.rastergis.populateStats(outMskImg, True, True, True)
        
        idx = 1
        refLyrsLst = []
        for img in inImages:
            print('In Image ('+str(idx) + '):\t' + img)
            baseImgName = os.path.splitext(os.path.basename(img))[0]
            refLyrNDVIImg = os.path.join(refLayersPath, baseImgName+'_ndvi.kea')
            refLyrNDWIImg = os.path.join(refLayersPath, baseImgName+'_ndwi.kea')
            refLyrLclMskImg = os.path.join(refLayersPath, baseImgName+'_waterLandMsk.kea')
            
            refLyrImg = os.path.join(refLayersPath, baseImgName+'_refHybrid.kea')
            bandDefns = []
            bandDefns.append(rsgislib.imagecalc.BandDefn('lmsk', refLyrLclMskImg, 1))
            bandDefns.append(rsgislib.imagecalc.BandDefn('omsk', outMskImg, 1))
            bandDefns.append(rsgislib.imagecalc.BandDefn('ndvi', refLyrNDVIImg, 1))
            bandDefns.append(rsgislib.imagecalc.BandDefn('ndwi', refLyrNDWIImg, 1))
            rsgislib.imagecalc.bandMath(refLyrImg, 'lmsk==0?-999:omsk==1?ndvi:omsk==2?ndwi:-999', 'KEA', rsgislib.TYPE_32FLOAT, bandDefns)
            refLyrsLst.append(refLyrImg)
            idx = idx + 1
        
        if incLS8 and (not onlyLS8):
            inImagesTmp = []
            for img in inImages:
                if 'LS8' in img:
                    baseImgName = os.path.splitext(os.path.basename(img))[0]
                    reflSubImg = os.path.join(tmpPath, baseImgName+'_bandsub.kea')
                    rsgislib.imageutils.selectImageBands(img, reflSubImg, 'KEA', dataType, [2,3,4,5,6,7])
                    inImagesTmp.append(reflSubImg)
                else:
                    inImagesTmp.append(img)
            inImages = inImagesTmp
        
        # Create REF Image
        rsgislib.imagecalc.getImgIdxForStat(refLyrsLst, outRefImg, 'KEA', -999, rsgislib.SUMTYPE_MAX)
        if calcStats:
            # Pop Ref Image with stats
            rsgislib.rastergis.populateStats(outRefImg, True, True, True)
            
            # Open the clumps dataset as a gdal dataset
            ratDataset = osgeo.gdal.Open(outRefImg, osgeo.gdal.GA_Update)
            # Write colours to RAT
            rios.rat.writeColumn(ratDataset, "Red", red)
            rios.rat.writeColumn(ratDataset, "Green", green)
            rios.rat.writeColumn(ratDataset, "Blue", blue)
            rios.rat.writeColumn(ratDataset, "Alpha", alpha)
            rios.rat.writeColumn(ratDataset, "Image", imgLyrs)
            ratDataset = None
        
        # Create Composite Image
        rsgislib.imageutils.createRefImgCompositeImg(inImages, outCompImg, outRefImg, gdalformat, dataType, 0.0)
        
        if calcStats:
            # Calc Stats
            rsgislib.imageutils.popImageStats(outCompImg, usenodataval=True, nodataval=0, calcpyramids=True)
        
        if not refImgTmpPresent:
            shutil.rmtree(refLayersPath, ignore_errors=True)
    
        if not tmpPresent:
            shutil.rmtree(tmpPath, ignore_errors=True)

    elif len(inImages) == 1:
        print("Only 1 Input Image, Just Copying File to output")
        shutil.copy(inImages[0], outCompImg)
    else:
        raise rsgislib.RSGISPyException("There were no input images for " + inImgsPattern)


def checkBuildLS8VRTs(in_imgs, outdir):
    """
A function which checks for Landsat 8 (LS8) images in the input list and creates
a band subset (i.e., removes the coastal (1) band). This function should be used
ahead of createMaxNDVINDWIComposite when combining LS8 with earlier landsat (i.e., 
LS5 and LS7) data. Note, files who's file name contain 'LS8' are considered to be 
Landsat 8 images.

:param in_imgs: input list of image file paths.
:param outdir: output directory for the VRT images (Note. an absolute path to the input image is used when building the VRT.)

"""
    import subprocess
    out_imgs = []
    for img in in_imgs:
        if 'LS8' in os.path.basename(img):
            abs_img = os.path.abspath(img)
            vrt_img = os.path.join(outdir, os.path.splitext(os.path.basename(img))[0]+'_bsub.vrt')
            cmd = "gdalbuildvrt -b 2 -b 3 -b 4 -b 5 -b 6 -b 7 {0} {1}".format(vrt_img, abs_img)
            subprocess.check_call(cmd, shell=True)
            out_imgs.append(vrt_img)
        else:
            out_imgs.append(img)
    return out_imgs


def createMaxNDVINDWIComposite(refImg, inImages, rBand, nBand, sBand, outRefImg, outCompImg, outMskImg, tmpPath='./tmp', gdalformat='KEA', dataType=None, calcStats=True, reprojmethod='cubic', use_mode=True):
    """
Create an image composite from multiple input images where the pixel brought through into the composite is the one with
the maximum NDVI over land and NDWI over water. A mask of land and water regions is also produced. The reference image is
used to define the spatial extent of the output images and spatial projection.

:param refImg: is a reference image with any number of bands and data type which is used to define the output image extent and projection.
:param inImages: a list of all the input images being used to build the composite.
:param rBand: is an integer specifying the red band in the input images (starts at 1), used in the NDVI index.
:param nBand: is an integer specifying the NIR band in the input images (starts at 1), used in the NDVI and NDWI index.
:param sBand: is an integer specifying the SWIR band in the input images (starts at 1), used in the NDVI and NDWI index.
:param outRefImg: is the output reference image which details which input image is forming the output image pixel value (Note. this file will always be a KEA file as RAT is used).
:param outCompImg: is the output composite image for which gdalformat and dataType define the format and data type.
:param outMskImg: is the output mask image for regions of water and land where ndvi vs ndwi are used (0=nodata, 1=land, 2=water)
:param tmpPath: is a temp path for intemediate files, if this path doesn't exist is will be created and deleted at runtime.
:param gdalformat: is the output file format of the outCompImg, any GDAL compatable format is OK (Defaut is KEA).
:param dataType: is the data type of the output image (outCompImg). If None is provided then the data type of the first input image will be used (Default None).
:param calcStats: calculate image statics and pyramids (Default=True)
:param reprojmethod : specifies the interpolation method used to reproject the input images which are in a different projection and/or pixel size as the reference image (default: cubic).
:param use_mode: True: the land/water masks are combined using the mode and False: the land water masks are combined using median.

    """
    rsgisUtils = rsgislib.RSGISPyUtils()
    uidStr = rsgisUtils.uidGenerator()
    
    if len(inImages) > 1:
        init_in_images = inImages
        inImagesOverlap = []
        inImages2ReProj = []
        inImagesSub2Ref = []
        nBands = 0
        first = True
        for img in init_in_images:
            if first:
                nBands = rsgisUtils.getImageBandCount(img)
                first = False
            else:
                cBands = rsgisUtils.getImageBandCount(img)
                if cBands != nBands:
                    raise rsgislib.RSGISPyException("The number of image bands is not consistent (Bands: {0} and {1})".format(nBands, cBands))

            sameProj = rsgisUtils.doGDALLayersHaveSameProj(refImg, img)
            if rsgislib.imageutils.doImagesOverlap(refImg, img):
                if sameProj:
                    if rsgisUtils.doImageResMatch(refImg, img):
                        inImagesOverlap.append(img)
                    else:
                        inImages2ReProj.append(img)
                else:
                    inImages2ReProj.append(img)
                    
        nImgs = len(inImagesOverlap) + len(inImages2ReProj)
        print("There are {} images with overlap to create the composite.".format(nImgs))
        if nImgs > 1:
            if dataType is None:
                if len(inImagesOverlap) > 0:
                    dataType = rsgisUtils.getRSGISLibDataTypeFromImg(inImagesOverlap[0])
                else:
                    dataType = rsgisUtils.getRSGISLibDataTypeFromImg(inImages2ReProj[0])
        
            tmpPresent = True
            if not os.path.exists(tmpPath):
                os.makedirs(tmpPath)
                tmpPresent = False 
            
            subsetImgTmpPresent = True
            if len(inImagesOverlap) > 0:
                print("Subset images to reference extent.")
                subsetLayersPath = os.path.join(tmpPath, 'Subset_'+uidStr)
                if not os.path.exists(subsetLayersPath):
                    os.makedirs(subsetLayersPath)
                    subsetImgTmpPresent = False
                
                for img in inImagesOverlap:
                    basename = os.path.splitext(os.path.basename(img))[0]
                    outImg = os.path.join(subsetLayersPath, basename+"_subset.kea")
                    rsgislib.imageutils.createCopyImage(refImg, outImg, nBands, 0, 'KEA', dataType)
                    rsgislib.imageutils.includeImagesIndImgIntersect(outImg, [img])
                    inImagesSub2Ref.append(outImg)
            
            reprojImgTmpPresent = True
            if len(inImages2ReProj) > 0:
                print("Reproject images to reference extent.")
                reprojLayersPath = os.path.join(tmpPath, 'Reproj_'+uidStr)
                if not os.path.exists(reprojLayersPath):
                    os.makedirs(reprojLayersPath)
                    reprojImgTmpPresent = False
                            
                for img in inImages2ReProj:
                    basename = os.path.splitext(os.path.basename(img))[0]
                    outImg = os.path.join(reprojLayersPath, basename+"_reproj.kea")
                    rsgislib.imageutils.resampleImage2Match(refImg, img, outImg, 'KEA', reprojmethod, dataType)
                    inImagesSub2Ref.append(outImg)
            
            refLayersPath = os.path.join(tmpPath, 'RefLyrs_'+uidStr)
            refImgTmpPresent = True
            if not os.path.exists(refLayersPath):
                os.makedirs(refLayersPath)
                refImgTmpPresent = False
                        
            numInLyrs = len(inImagesSub2Ref)
            
            numpy.random.seed(5)
            red = numpy.random.randint(1, 255, numInLyrs+1, int)
            red[0] = 0
            numpy.random.seed(2)
            green = numpy.random.randint(1, 255, numInLyrs+1, int)
            green[0] = 0
            numpy.random.seed(9)
            blue = numpy.random.randint(1, 255, numInLyrs+1, int)
            blue[0] = 0
            alpha = numpy.zeros_like(blue)
            alpha[...] = 255
            imgLyrs = numpy.empty(numInLyrs+1, dtype=numpy.dtype('a255'))
                
            # Generate Comp Ref layers:
            landWaterMskBandDefns = []
            mskImgs = []
            idx = 1
            for img in inImagesSub2Ref:
                print('In Image ('+str(idx) + '):\t' + img)
                imgLyrs[idx] = os.path.basename(img)
                baseImgName = os.path.splitext(os.path.basename(img))[0]
                refLyrNDVIImg = os.path.join(refLayersPath, baseImgName+'_ndvi.kea')
                refLyrNDWIImg = os.path.join(refLayersPath, baseImgName+'_ndwi.kea')
                rsgislib.imagecalc.calcindices.calcNDVI(img, rBand, nBand, refLyrNDVIImg, False)
                rsgislib.imagecalc.calcindices.calcNDWI(img, nBand, sBand, refLyrNDWIImg, False)
                
                refLyrMskImg = os.path.join(refLayersPath, baseImgName+'_waterLandMsk.kea')
                bandDefns = []
                bandDefns.append(rsgislib.imagecalc.BandDefn('ndvi', refLyrNDVIImg, 1))
                bandDefns.append(rsgislib.imagecalc.BandDefn('ndwi', refLyrNDWIImg, 1))
                rsgislib.imagecalc.bandMath(refLyrMskImg, 'ndvi<-1?0:ndvi>0.3?1:ndwi>0.01?2:1', 'KEA', rsgislib.TYPE_8UINT, bandDefns)
                mskImgs.append(refLyrMskImg)
                idx = idx + 1
            imgLyrs[0] = ""        
            
            refLyrMskStackImg = os.path.join(refLayersPath, uidStr+'_waterLandMskStack.kea')
            rsgislib.imageutils.stackImageBands(mskImgs, None, refLyrMskStackImg, -1, -1, 'KEA', rsgislib.TYPE_8UINT)
            if use_mode:
                rsgislib.imagecalc.imagePixelColumnSummary(refLyrMskStackImg, outMskImg,
                                                           rsgislib.imagecalc.StatsSummary(calcMode=True), 'KEA',
                                                           rsgislib.TYPE_8UINT, 0, True)
            else:
                rsgislib.imagecalc.imagePixelColumnSummary(refLyrMskStackImg, outMskImg,
                                                           rsgislib.imagecalc.StatsSummary(calcMedian=True), 'KEA',
                                                           rsgislib.TYPE_8UINT, 0, True)
            rsgislib.rastergis.populateStats(outMskImg, True, True, True)
            
            idx = 1
            refLyrsLst = []
            for img in inImagesSub2Ref:
                print('In Image ('+str(idx) + '):\t' + img)
                baseImgName = os.path.splitext(os.path.basename(img))[0]
                refLyrNDVIImg = os.path.join(refLayersPath, baseImgName+'_ndvi.kea')
                refLyrNDWIImg = os.path.join(refLayersPath, baseImgName+'_ndwi.kea')
                refLyrLclMskImg = os.path.join(refLayersPath, baseImgName+'_waterLandMsk.kea')
                
                refLyrImg = os.path.join(refLayersPath, baseImgName+'_refHybrid.kea')
                bandDefns = []
                bandDefns.append(rsgislib.imagecalc.BandDefn('lmsk', refLyrLclMskImg, 1))
                bandDefns.append(rsgislib.imagecalc.BandDefn('omsk', outMskImg, 1))
                bandDefns.append(rsgislib.imagecalc.BandDefn('ndvi', refLyrNDVIImg, 1))
                bandDefns.append(rsgislib.imagecalc.BandDefn('ndwi', refLyrNDWIImg, 1))
                rsgislib.imagecalc.bandMath(refLyrImg, 'lmsk==0?-999:omsk==1?ndvi:omsk==2?ndwi:-999', 'KEA', rsgislib.TYPE_32FLOAT, bandDefns)
                refLyrsLst.append(refLyrImg)
                idx = idx + 1
            
            # Create REF Image
            rsgislib.imagecalc.getImgIdxForStat(refLyrsLst, outRefImg, 'KEA', -999, rsgislib.SUMTYPE_MAX)
            if calcStats:
                # Pop Ref Image with stats
                rsgislib.rastergis.populateStats(outRefImg, True, True, True)
                
                # Open the clumps dataset as a gdal dataset
                ratDataset = osgeo.gdal.Open(outRefImg, osgeo.gdal.GA_Update)
                # Write colours to RAT
                rios.rat.writeColumn(ratDataset, "Red", red)
                rios.rat.writeColumn(ratDataset, "Green", green)
                rios.rat.writeColumn(ratDataset, "Blue", blue)
                rios.rat.writeColumn(ratDataset, "Alpha", alpha)
                rios.rat.writeColumn(ratDataset, "Image", imgLyrs)
                ratDataset = None
            
            # Create Composite Image
            rsgislib.imageutils.createRefImgCompositeImg(inImagesSub2Ref, outCompImg, outRefImg, gdalformat, dataType, 0.0)
            
            if calcStats:
                # Calc Stats
                rsgislib.imageutils.popImageStats(outCompImg, usenodataval=True, nodataval=0, calcpyramids=True)
            
            if not refImgTmpPresent:
                shutil.rmtree(refLayersPath, ignore_errors=True)
                
            if not subsetImgTmpPresent:
                shutil.rmtree(subsetLayersPath, ignore_errors=True)    
                
            if not reprojImgTmpPresent:
                shutil.rmtree(reprojLayersPath, ignore_errors=True)            
        
            if not tmpPresent:
                shutil.rmtree(tmpPath, ignore_errors=True)
        else:
            raise rsgislib.RSGISPyException("There were only {0} images which intersect with the reference image.".format(nImgs))

    elif len(inImages) == 1:
        print("Only 1 Input Image, Just Copying File to output")
        nBands = rsgisUtils.getImageBandCount(inImages[0])
        if dataType is None:
            dataType = rsgisUtils.getRSGISLibDataTypeFromImg(inImages[0])
        rsgislib.imageutils.createCopyImage(refImg, outCompImg, nBands, 0, gdalformat, dataType)
        rsgislib.imageutils.includeImagesIndImgIntersect(outCompImg, [inImages[0]])
    else:
        raise rsgislib.RSGISPyException("List of input images was empty ")







#! /usr/bin/env python

############################################################################
#  pxl_outlier_chng.py
#
#  Copyright 2019 RSGISLib.
#
#  RSGISLib: 'The remote sensing and GIS Software Library'
#
#  RSGISLib is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  RSGISLib is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with RSGISLib.  If not, see <http://www.gnu.org/licenses/>.
#
#
# Purpose:  Provide a set of utilities to perform per-pixel change detection
#           using outlier methods. These are sometime called map-to-image 
#           methods of change detection.
#
# Author: Pete Bunting
# Email: petebunting@mac.com
# Date: 12/06/2019
# Version: 1.0
#
# History:
# Version 1.0 - Created.
#
###########################################################################

from __future__ import print_function

import numpy

import rsgislib
import rsgislib.rastergis
import rsgislib.imageutils

haveRIOS = True
try:
    from rios import applier
    from rios import cuiprogress
    from rios import rat
except ImportError as riosErr:
    haveRIOS = False


def find_class_outliers(pyod_obj, img, img_mask, out_lbls_img, out_scores_img=None, img_mask_val=1, img_bands=None,
                        gdalformat="KEA"):
    """
This function uses the pyod (https://github.com/yzhao062/pyod) library to find outliers within a class.
It is assumed that the input images are from a different date than the mask (classification) and therefore
the outliners will related to class changes.

:param pyod_obj: an instance of a pyod.models (e.g., pyod.models.knn.KNN) pass parameters to the constructor
:param img: input image used for analysis
:param img_mask: input image mask use to define the region of interest.
:param out_lbls_img: output image with pixel over of 1 for within mask but not outlier and 2 for in mask and outlier.
:param out_scores_img: output image (optional, None and won't be provided; Default None) providing the probability of
                       each pixel being an outlier
:param img_mask_val: the pixel value within the mask image for the class of interest. (Default 1)
:param img_bands: the image bands to be used for the analysis. If None then all used (Default: None)
:param gdalformat: file format for the output image(s). Default KEA.

"""
    if not haveRIOS:
        raise Exception("The rios module is required for this function could not be imported\n\t" + riosErr)

    rsgis_utils = rsgislib.RSGISPyUtils()

    if img_bands is not None:
        if not ((type(img_bands) is list) or (type(img_bands) is tuple)):
            raise rsgislib.RSGISPyException("If provided then img_bands should be a list (or None)")
    else:
        n_bands = rsgis_utils.getImageBandCount(img)
        img_bands = numpy.arange(1, n_bands + 1)
    num_vars = len(img_bands)
    img_val_no_data = rsgis_utils.getImageNoDataValue(img)

    msk_arr_vals = rsgislib.imageutils.extractImgPxlValsInMsk(img, img_bands, img_mask, img_mask_val, img_val_no_data)
    print("There were {} pixels within the mask.".format(msk_arr_vals.shape[0]))

    print("Fitting oulier detector")
    pyod_obj.fit(msk_arr_vals)
    print("Fitted oulier detector")

    # RIOS function to apply classifer
    def _applyPyOB(info, inputs, outputs, otherargs):
        # Internal function for rios applier. Used within find_class_outliers.

        out_lbls_vals = numpy.zeros_like(inputs.image_mask, dtype=numpy.uint8)
        if otherargs.out_scores:
            out_scores_vals = numpy.zeros_like(inputs.image_mask, dtype=numpy.float)
        if numpy.any(inputs.image_mask == otherargs.msk_val):
            out_lbls_vals = out_lbls_vals.flatten()
            img_msk_vals = inputs.image_mask.flatten()
            if otherargs.out_scores:
                out_scores_vals = out_scores_vals.flatten()
            ID = numpy.arange(img_msk_vals.shape[0])

            img_shape = inputs.img.shape
            img_bands = inputs.img.reshape((img_shape[0], (img_shape[1] * img_shape[2])))

            band_lst = []
            for band in otherargs.img_bands:
                if (band > 0) and (band <= img_shape[0]):
                    band_lst.append(img_bands[band - 1])
                else:
                    raise Exception("Band ({}) specified is not within the image".format(band))
            img_bands_sel = numpy.stack(band_lst, axis=0)
            img_bands_trans = numpy.transpose(img_bands_sel)

            if otherargs.no_data_val is not None:
                ID = ID[(img_bands_trans != otherargs.no_data_val).all(axis=1)]
                img_msk_vals = img_msk_vals[(img_bands_trans != otherargs.no_data_val).all(axis=1)]
                img_bands_trans = img_bands_trans[(img_bands_trans != otherargs.no_data_val).all(axis=1)]

            ID = ID[img_msk_vals == otherargs.msk_val]
            img_bands_trans = img_bands_trans[img_msk_vals == otherargs.msk_val]

            if img_bands_trans.shape[0] > 0:
                pred_lbls = otherargs.pyod_obj.predict(img_bands_trans)
                pred_lbls = pred_lbls + 1
                out_lbls_vals[ID] = pred_lbls
            out_lbls_vals = numpy.expand_dims(
                out_lbls_vals.reshape((inputs.image_mask.shape[1], inputs.image_mask.shape[2])), axis=0)

            if otherargs.out_scores:
                if img_bands_trans.shape[0] > 0:
                    pred_probs = otherargs.pyod_obj.predict_proba(img_bands_trans, method='unify')
                    out_scores_vals[ID] = pred_probs[:, 1]
                out_scores_vals = numpy.expand_dims(
                    out_scores_vals.reshape((inputs.image_mask.shape[1], inputs.image_mask.shape[2])), axis=0)

        outputs.out_lbls_img = out_lbls_vals
        if otherargs.out_scores:
            outputs.out_scores_img = out_scores_vals

    infiles = applier.FilenameAssociations()
    infiles.image_mask = img_mask
    infiles.img = img

    otherargs = applier.OtherInputs()
    otherargs.pyod_obj = pyod_obj
    otherargs.msk_val = img_mask_val
    otherargs.img_bands = img_bands
    otherargs.out_scores = False
    otherargs.no_data_val = img_val_no_data

    outfiles = applier.FilenameAssociations()
    outfiles.out_lbls_img = out_lbls_img
    if out_scores_img is not None:
        outfiles.out_scores_img = out_scores_img
        otherargs.out_scores = True

    aControls = applier.ApplierControls()
    aControls.progress = cuiprogress.CUIProgressBar()
    aControls.drivername = gdalformat
    aControls.omitPyramids = True
    aControls.calcStats = False
    print("Applying the Outlier Detector")
    applier.apply(_applyPyOB, infiles, outfiles, otherargs, controls=aControls)
    print("Completed")

    rsgislib.rastergis.populateStats(clumps=out_lbls_img, addclrtab=True, calcpyramids=True, ignorezero=True)
    if out_scores_img is not None:
        rsgislib.imageutils.popImageStats(out_scores_img, usenodataval=True, nodataval=0, calcpyramids=True)



#! /usr/bin/env python
"""
The tools.utm module contains some useful tools for work with UTM projection.

This code has been adapted from https://github.com/Turbo87/utm/blob/master/utm/conversion.py

"""
import numpy

K0 = 0.9996

E = 0.00669438
E2 = E * E
E3 = E2 * E
E_P2 = E / (1.0 - E)

SQRT_E = numpy.sqrt(1 - E)
_E = (1 - SQRT_E) / (1 + SQRT_E)
_E2 = _E * _E
_E3 = _E2 * _E
_E4 = _E3 * _E
_E5 = _E4 * _E

M1 = (1 - E / 4 - 3 * E2 / 64 - 5 * E3 / 256)
M2 = (3 * E / 8 + 3 * E2 / 32 + 45 * E3 / 1024)
M3 = (15 * E2 / 256 + 45 * E3 / 1024)
M4 = (35 * E3 / 3072)

P2 = (3. / 2 * _E - 27. / 32 * _E3 + 269. / 512 * _E5)
P3 = (21. / 16 * _E2 - 55. / 32 * _E4)
P4 = (151. / 96 * _E3 - 417. / 128 * _E5)
P5 = (1097. / 512 * _E4)

R = 6378137

ZONE_LETTERS = "CDEFGHJKLMNPQRSTUVWXX"


def min_max_eastings_for_lat(latitude, zone=1):
    """
Find the minimum and maximum Eastings for a specified latitude within a specific zone.
This should be the same for all zones.

:param latitude: float
:param zone: int
:return: tuple [min, max]

"""
    if isinstance(latitude, numpy.ndarray):
        raise Exception("Don't input an array.")

    central_long = zone_number_to_central_longitude(zone)
    min_long = central_long - 3.0
    max_long = central_long + 3.0

    min_utm = from_latlon(latitude, min_long, force_zone_number=zone)
    max_utm = from_latlon(latitude, max_long, force_zone_number=zone)

    return min_utm[0], max_utm[0]


def epsg_for_UTM(zone, hemisphere):
    """
Return EPSG code for given UTM zone and hemisphere using WGS84 datum.

:param zone: UTM zone
:param hemisphere: hemisphere either 'N' or 'S'
:return: corresponding EPSG code

"""
    if hemisphere not in ['N', 'S']:
        raise Exception('Invalid hemisphere ("N" or "S").')

    if zone < 0 or zone > 60:
        raise Exception('UTM zone ouside valid range.')

    if hemisphere == 'N':
        ns = 600
    else:
        ns = 700

    if zone == 0:
        zone = 61

    return int(32000 + ns + zone)


def split_wgs84_bbox_utm_zones(wgs84_bbox):
    """
Function which splits a bounding box (MinX, MaxX, MinY, MaxY) into
a number of small bounding boxes (MinX, MaxX, MinY, MaxY) at UTM zone
boundaries.

:param wgs84_bbox: bounding box in lat/long WGS84 (MinX, MaxX, MinY, MaxY)
:return: list [utm zone, bbox]

"""
    utm_tl = from_latlon(wgs84_bbox[3], wgs84_bbox[0])
    utm_tr = from_latlon(wgs84_bbox[3], wgs84_bbox[1])
    utm_br = from_latlon(wgs84_bbox[2], wgs84_bbox[1])
    utm_bl = from_latlon(wgs84_bbox[2], wgs84_bbox[0])

    utm_zones = []
    for utm_coord in [utm_tl, utm_tr, utm_br, utm_bl]:
        utm_zones.append(utm_coord[2])

    min_zone = numpy.min(utm_zones)
    max_zone = numpy.max(utm_zones)

    zones = numpy.arange(min_zone, max_zone+1, 1)

    out_bboxs = []
    for zone in zones:
        zone_min = ((zone*6.0)-6.0)-180.0
        zone_max = (zone*6.0)-180.0

        # Intersect input BBOX with zone bound
        x_min = zone_min
        if wgs84_bbox[0] > zone_min:
            x_min = wgs84_bbox[0]

        x_max = zone_max
        if wgs84_bbox[1] < zone_max:
            x_max = wgs84_bbox[1]

        out_bboxs.append([zone, [x_min, x_max, wgs84_bbox[2], wgs84_bbox[3]]])

    return out_bboxs


def check_valid_zone(zone_number, zone_letter):
    """
    Check that the UTM zone is valid.

:param zone_number: int
:param zone_letter: string

:return: True if successful otherwise raises an exception.

"""
    if not 1 <= zone_number <= 60:
        raise Exception('zone number out of range (must be between 1 and 60)')

    if zone_letter:
        zone_letter = zone_letter.upper()

        if not 'C' <= zone_letter <= 'X' or zone_letter in ['I', 'O']:
            raise Exception('zone letter out of range (must be between C and X)')

    return True


def to_latlon(easting, northing, zone_number, zone_letter=None, northern=None, strict=True):
    """
This function convert an UTM coordinate into Latitude and Longitude

:param easting: int or array: Easting value of UTM coordinate
:param northing: int or array: Northing value of UTM coordinate
:param zone number: int: Zone Number is represented with global map numbers of an UTM Zone
    Numbers Map. More information see http://www.jaworski.ca/utmzones.htm
:param zone_letter: str: Zone Letter can be represented as string values.
    Where UTM Zone Designators can be accessed in http://www.jaworski.ca/utmzones.htm
:param northern: bool: You can set True or False to set this parameter. Default is None

"""
    import rsgislib
    rsgis_utils = rsgislib.RSGISPyUtils()
    if not zone_letter and northern is None:
        raise ValueError('either zone_letter or northern needs to be set')

    elif zone_letter and northern is not None:
        raise ValueError('set either zone_letter or northern, but not both')

    if strict:
        if not rsgis_utils.in_bounds(easting, 100000, 1000000, upper_strict=True):
            raise Exception('easting out of range (must be between 100.000 m and 999.999 m)')
        if not rsgis_utils.in_bounds(northing, 0, 10000000):
            raise Exception('northing out of range (must be between 0 m and 10.000.000 m)')

    check_valid_zone(zone_number, zone_letter)

    if zone_letter:
        zone_letter = zone_letter.upper()
        northern = (zone_letter >= 'N')

    x = easting - 500000
    y = northing

    if not northern:
        y -= 10000000

    m = y / K0
    mu = m / (R * M1)

    p_rad = (mu +
             P2 * numpy.sin(2 * mu) +
             P3 * numpy.sin(4 * mu) +
             P4 * numpy.sin(6 * mu) +
             P5 * numpy.sin(8 * mu))

    p_sin = numpy.sin(p_rad)
    p_sin2 = p_sin * p_sin

    p_cos = numpy.cos(p_rad)

    p_tan = p_sin / p_cos
    p_tan2 = p_tan * p_tan
    p_tan4 = p_tan2 * p_tan2

    ep_sin = 1 - E * p_sin2
    ep_sin_sqrt = numpy.sqrt(1 - E * p_sin2)

    n = R / ep_sin_sqrt
    r = (1 - E) / ep_sin

    c = _E * p_cos ** 2
    c2 = c * c

    d = x / (n * K0)
    d2 = d * d
    d3 = d2 * d
    d4 = d3 * d
    d5 = d4 * d
    d6 = d5 * d

    latitude = (p_rad - (p_tan / r) *
                (d2 / 2 -
                 d4 / 24 * (5 + 3 * p_tan2 + 10 * c - 4 * c2 - 9 * E_P2)) +
                d6 / 720 * (61 + 90 * p_tan2 + 298 * c + 45 * p_tan4 - 252 * E_P2 - 3 * c2))

    longitude = (d -
                 d3 / 6 * (1 + 2 * p_tan2 + c) +
                 d5 / 120 * (5 - 2 * c + 28 * p_tan2 - 3 * c2 + 8 * E_P2 + 24 * p_tan4)) / p_cos

    return (numpy.degrees(latitude),
            numpy.degrees(longitude) + zone_number_to_central_longitude(zone_number))


def from_latlon(latitude, longitude, force_zone_number=None, force_zone_letter=None):
    """
This function convert Latitude and Longitude to UTM coordinate

:param latitude: float or array. Latitude between 80 deg S and 84 deg N, e.g. (-80.0 to 84.0)
:param longitude: float or array: Longitude between 180 deg W and 180 deg E, e.g. (-180.0 to 180.0).
:param force_zone number: int: Zone Number is represented with global map numbers of an UTM Zone
    Numbers Map. You may force conversion including one UTM Zone Number.
    More information see http://www.jaworski.ca/utmzones.htm

"""
    import rsgislib
    rsgis_utils = rsgislib.RSGISPyUtils()
    if not rsgis_utils.in_bounds(latitude, -80.0, 84.0):
        raise Exception('latitude out of range (must be between 80 deg S and 84 deg N)')
    if not rsgis_utils.in_bounds(longitude, -180.0, 180.0):
        raise Exception('longitude out of range (must be between 180 deg W and 180 deg E)')
    if force_zone_number is not None:
        check_valid_zone(force_zone_number, force_zone_letter)

    lat_rad = numpy.radians(latitude)
    lat_sin = numpy.sin(lat_rad)
    lat_cos = numpy.cos(lat_rad)

    lat_tan = lat_sin / lat_cos
    lat_tan2 = lat_tan * lat_tan
    lat_tan4 = lat_tan2 * lat_tan2

    if force_zone_number is None:
        zone_number = latlon_to_zone_number(latitude, longitude)
    else:
        zone_number = force_zone_number

    if force_zone_letter is None:
        zone_letter = latitude_to_zone_letter(latitude)
    else:
        zone_letter = force_zone_letter

    lon_rad = numpy.radians(longitude)
    central_lon = zone_number_to_central_longitude(zone_number)
    central_lon_rad = numpy.radians(central_lon)

    n = R / numpy.sqrt(1 - E * lat_sin ** 2)
    c = E_P2 * lat_cos ** 2

    a = lat_cos * (lon_rad - central_lon_rad)
    a2 = a * a
    a3 = a2 * a
    a4 = a3 * a
    a5 = a4 * a
    a6 = a5 * a

    m = R * (M1 * lat_rad -
             M2 * numpy.sin(2 * lat_rad) +
             M3 * numpy.sin(4 * lat_rad) -
             M4 * numpy.sin(6 * lat_rad))

    easting = K0 * n * (a +
                        a3 / 6 * (1 - lat_tan2 + c) +
                        a5 / 120 * (5 - 18 * lat_tan2 + lat_tan4 + 72 * c - 58 * E_P2)) + 500000

    northing = K0 * (m + n * lat_tan * (a2 / 2 +
                                        a4 / 24 * (5 - lat_tan2 + 9 * c + 4 * c ** 2) +
                                        a6 / 720 * (61 - 58 * lat_tan2 + lat_tan4 + 600 * c - 330 * E_P2)))

    if rsgis_utils.mixed_signs(latitude):
        raise ValueError("latitudes must all have the same sign")
    elif rsgis_utils.negative(latitude):
        northing += 10000000

    return easting, northing, zone_number, zone_letter


def latitude_to_zone_letter(latitude):
    """
Find the zone letter for the latitude. If the input is a numpy array, just use the first element
user responsibility to make sure that all points are in one zone

:param latitude:
:return: string if found otherwise None.

"""
    if isinstance(latitude, numpy.ndarray):
        latitude = latitude.flat[0]

    if -80 <= latitude <= 84:
        return ZONE_LETTERS[int(latitude + 80) >> 3]
    else:
        return None


def latlon_to_zone_number(latitude, longitude):
    """
Find the UTM zone number for a give latitude and longitude. If the input is a numpy array, just use the
first element user responsibility to make sure that all points are in one zone

:param latitude: float
:param longitude: float

:return: int

"""
    if isinstance(latitude, numpy.ndarray):
        latitude = latitude.flat[0]
    if isinstance(longitude, numpy.ndarray):
        longitude = longitude.flat[0]

    if 56 <= latitude < 64 and 3 <= longitude < 12:
        return 32

    if 72 <= latitude <= 84 and longitude >= 0:
        if longitude < 9:
            return 31
        elif longitude < 21:
            return 33
        elif longitude < 33:
            return 35
        elif longitude < 42:
            return 37

    return int((longitude + 180) / 6) + 1


def zone_number_to_central_longitude(zone_number):
    """
Find the central longitude for the given zone.

:param zone_number: int
:return: float

"""
    return (zone_number - 1) * 6 - 180 + 3



#! /usr/bin/env python

############################################################################
#  tiledclump.py
#
#  Copyright 2016 RSGISLib.
#
#  RSGISLib: 'The Remote Sensing and GIS Software Library'
#
#  RSGISLib is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  RSGISLib is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with RSGISLib.  If not, see <http://www.gnu.org/licenses/>.
#
#
# Purpose:  Provide functionality to clump an image using image tiles.
#
# Author: Pete Bunting
# Email: petebunting@mac.com
# Date: 21/3/2016
# Version: 1.0
#
# History:
# Version 1.0 - Created.
#
###########################################################################

import glob
import os.path
import os
import shutil
from multiprocessing import Pool
import multiprocessing

import rsgislib
from rsgislib import segmentation
from rsgislib import imageutils

################################ Clumping Functions ################################

def performClumpingSingleThread(inputImage, clumpsImage, tmpDIR='tmp', width=2000, height=2000, gdalformat='KEA'):
    """
Clump the input image using a tiled processing chain allowing large images to be clumped more quickly.

:param inputImage: the input image to be clumped.
:param clumpsImage: the output clumped image.
:param tmpDIR: the temporary directory where intermediate files will be written (default is 'tmp'). Directory will be created and deleted if does not exist.
:param width: int for width of the image tiles used for processing (Default = 2000).
:param height: int for height of the image tiles used for processing (Default = 2000).
:param gdalformat: string with the GDAL image format for the output image (Default = KEA). NOTE. KEA is used as intermediate format internally and therefore needs to be available.

    """
    createdTmp = False
    if not os.path.exists(tmpDIR):
        os.makedirs(tmpDIR)
        createdTmp = True
    
    rsgisUtils = rsgislib.RSGISPyUtils()
    uidStr = rsgisUtils.uidGenerator()
    dataType = rsgisUtils.getRSGISLibDataTypeFromImg(inputImage)
    baseName = os.path.splitext(os.path.basename(inputImage))[0]+"_"+uidStr
    imgTilesDIR = os.path.join(tmpDIR, "imgtiles_"+uidStr)
    tilesClumpsDIR = os.path.join(tmpDIR, "imgclumpstiles_"+uidStr)
    tilesImgBase = os.path.join(imgTilesDIR, baseName)
    initMergedClumps = os.path.join(tmpDIR, "MergedInitClumps_"+uidStr+".kea")
    if not os.path.exists(imgTilesDIR):
        os.makedirs(imgTilesDIR)
    if not os.path.exists(tilesClumpsDIR):
        os.makedirs(tilesClumpsDIR)
    
    imageutils.createTiles(inputImage, tilesImgBase, int(width), int(height), 0, False, 'KEA', dataType, 'kea')
    imageTiles = glob.glob(tilesImgBase+"*")

    for tile in imageTiles:
        tilBaseName = os.path.splitext(os.path.basename(tile))[0]
        clumpedTile = os.path.join(tilesClumpsDIR, tilBaseName+'_clumps.kea')
        segmentation.clump(tile, clumpedTile, 'KEA', True, 0, True)
    
    clumpTiles = glob.glob(os.path.join(tilesClumpsDIR, '*_clumps.kea'))
    print("Create Blank Image")
    imageutils.createCopyImage(inputImage, initMergedClumps, 1, 0, 'KEA', rsgislib.TYPE_32UINT)
    print("Merge Tiles into Blank Image")
    segmentation.mergeClumpImages(clumpTiles, initMergedClumps, True)
    print("Merge Tile Boundaries")
    segmentation.mergeEquivClumps(initMergedClumps, clumpsImage, gdalformat, ['PixelVal'])
    
    shutil.rmtree(imgTilesDIR)
    shutil.rmtree(tilesClumpsDIR)
    os.remove(initMergedClumps)
    if createdTmp:
        shutil.rmtree(tmpDIR)


def clumpImgFunc(imgs):
    """
    Clump an image with values provides as an array for use within a multiprocessing Pool
    """
    segmentation.clump(imgs[0], imgs[1], 'KEA', True, 0, True)


def performClumpingMultiProcess(inputImage, clumpsImage, tmpDIR='tmp', width=2000, height=2000, gdalformat='KEA', nCores=-1):
    """
Clump the input image using a tiled processing chain allowing large images to be clumped more quickly.

:param inputImage: the input image to be clumped.
:param clumpsImage: the output clumped image.
:param tmpDIR: the temporary directory where intermediate files will be written (default is 'tmp'). Directory will be created and deleted if does not exist.
:param width: int for width of the image tiles used for processing (Default = 2000).
:param height: int for height of the image tiles used for processing (Default = 2000).
:param gdalformat: string with the GDAL image format for the output image (Default = KEA). NOTE. KEA is used as intermediate format internally and therefore needs to be available.
:param nCores: is an int specifying the number of cores to be used for clumping processing.

    """
    createdTmp = False
    if not os.path.exists(tmpDIR):
        os.makedirs(tmpDIR)
        createdTmp = True
    
    if nCores <= 0:
        nCores = multiprocessing.cpu_count()
    
    rsgisUtils = rsgislib.RSGISPyUtils()
    uidStr = rsgisUtils.uidGenerator()
    dataType = rsgisUtils.getRSGISLibDataTypeFromImg(inputImage)
    baseName = os.path.splitext(os.path.basename(inputImage))[0]+"_"+uidStr
    imgTilesDIR = os.path.join(tmpDIR, "imgtiles_"+uidStr)
    tilesClumpsDIR = os.path.join(tmpDIR, "imgclumpstiles_"+uidStr)
    tilesImgBase = os.path.join(imgTilesDIR, baseName)
    initMergedClumps = os.path.join(tmpDIR, "MergedInitClumps_"+uidStr+".kea")
    if not os.path.exists(imgTilesDIR):
        os.makedirs(imgTilesDIR)
    if not os.path.exists(tilesClumpsDIR):
        os.makedirs(tilesClumpsDIR)
    
    imageutils.createTiles(inputImage, tilesImgBase, int(width), int(height), 0, False, 'KEA', dataType, 'kea')
    imageTiles = glob.glob(tilesImgBase+"*")
    
    clumpImgsVals = []
    for tile in imageTiles:
        tilBaseName = os.path.splitext(os.path.basename(tile))[0]
        clumpedTile = os.path.join(tilesClumpsDIR, tilBaseName+'_clumps.kea')
        clumpImgsVals.append([tile, clumpedTile])
    
    with Pool(nCores) as p:
        p.map(clumpImgFunc, clumpImgsVals)
    
    clumpTiles = glob.glob(os.path.join(tilesClumpsDIR, '*_clumps.kea'))
    print("Create Blank Image")
    imageutils.createCopyImage(inputImage, initMergedClumps, 1, 0, 'KEA', rsgislib.TYPE_32UINT)
    print("Merge Tiles into Blank Image")
    segmentation.mergeClumpImages(clumpTiles, initMergedClumps, True)
    print("Merge Tile Boundaries")
    segmentation.mergeEquivClumps(initMergedClumps, clumpsImage, gdalformat, ['PixelVal'])
    
    shutil.rmtree(imgTilesDIR)
    shutil.rmtree(tilesClumpsDIR)
    os.remove(initMergedClumps)
    if createdTmp:
        shutil.rmtree(tmpDIR)
################################################################################################


################################ Union Clumping Functions ################################

def performUnionClumpingSingleThread(inputImage, refImg, clumpsImage, tmpDIR='tmp', width=2000, height=2000, gdalformat='KEA'):
    """
Clump and union with the reference image the input image using a tiled processing chain allowing large images to be clumped more quickly.

:param inputImage: the input image to be clumped.
:param refImg: the reference image which the union is undertaken with (typically an existing classification)
:param clumpsImage: the output clumped image.
:param tmpDIR: the temporary directory where intermediate files will be written (default is 'tmp'). Directory will be created and deleted if does not exist.
:param width: int for width of the image tiles used for processing (Default = 2000).
:param height: int for height of the image tiles used for processing (Default = 2000).
:param gdalformat: string with the GDAL image format for the output image (Default = KEA). NOTE. KEA is used as intermediate format internally and therefore needs to be available.

    """
    createdTmp = False
    if not os.path.exists(tmpDIR):
        os.makedirs(tmpDIR)
        createdTmp = True
    
    rsgisUtils = rsgislib.RSGISPyUtils()
    uidStr = rsgisUtils.uidGenerator()
    dataType = rsgisUtils.getRSGISLibDataTypeFromImg(inputImage)
    baseName = os.path.splitext(os.path.basename(inputImage))[0]+"_"+uidStr
    imgTilesDIR = os.path.join(tmpDIR, "imgtiles_"+uidStr)
    tilesClumpsDIR = os.path.join(tmpDIR, "imgclumpstiles_"+uidStr)
    tilesImgBase = os.path.join(imgTilesDIR, baseName)
    initMergedClumps = os.path.join(tmpDIR, "MergedInitClumps_"+uidStr+".kea")
    if not os.path.exists(imgTilesDIR):
        os.makedirs(imgTilesDIR)
    if not os.path.exists(tilesClumpsDIR):
        os.makedirs(tilesClumpsDIR)
    
    imageutils.createTiles(inputImage, tilesImgBase, int(width), int(height), 0, False, 'KEA', dataType, 'kea')
    imageTiles = glob.glob(tilesImgBase+"*")

    for tile in imageTiles:
        tilBaseName = os.path.splitext(os.path.basename(tile))[0]
        clumpedTile = os.path.join(tilesClumpsDIR, tilBaseName+'_clumps.kea')
        segmentation.unionOfClumps(clumpedTile, 'KEA', [tile, refImg], 0, True)
    
    clumpTiles = glob.glob(os.path.join(tilesClumpsDIR, '*_clumps.kea'))
    print("Create Blank Image")
    imageutils.createCopyImage(inputImage, initMergedClumps, 1, 0, 'KEA', rsgislib.TYPE_32UINT)
    print("Merge Tiles into Blank Image")
    segmentation.mergeClumpImages(clumpTiles, initMergedClumps, True)
    print("Merge Tile Boundaries")
    segmentation.mergeEquivClumps(initMergedClumps, clumpsImage, gdalformat, ['ClumpVal_1', 'ClumpVal_2'])
    
    shutil.rmtree(imgTilesDIR)
    shutil.rmtree(tilesClumpsDIR)
    os.remove(initMergedClumps)
    if createdTmp:
        shutil.rmtree(tmpDIR)

def unionClumpImgFunc(imgs):
    """
    Union Clump an image with values provides as an array for use within a multiprocessing Pool
    """
    segmentation.unionOfClumps(imgs[2], 'KEA', [imgs[0], imgs[1]], 0, True)
    
def performUnionClumpingMultiProcess(inputImage, refImg, clumpsImage, tmpDIR='tmp', width=2000, height=2000, gdalformat='KEA', nCores=-1):
    """
Clump and union with the reference image the input image using a tiled processing chain allowing large images to be clumped more quickly.

:param inputImage: the input image to be clumped.
:param refImg: the reference image which the union is undertaken with (typically an existing classification)
:param clumpsImage: the output clumped image.
:param tmpDIR: the temporary directory where intermediate files will be written (default is 'tmp'). Directory will be created and deleted if does not exist.
:param width: int for width of the image tiles used for processing (Default = 2000).
:param height: int for height of the image tiles used for processing (Default = 2000).
:param gdalformat: string with the GDAL image format for the output image (Default = KEA). NOTE. KEA is used as intermediate format internally and therefore needs to be available.
:param nCores: is an int specifying the number of cores to be used for clumping processing.

    """
    createdTmp = False
    if not os.path.exists(tmpDIR):
        os.makedirs(tmpDIR)
        createdTmp = True
    
    if nCores <= 0:
        nCores = multiprocessing.cpu_count()
    
    rsgisUtils = rsgislib.RSGISPyUtils()
    uidStr = rsgisUtils.uidGenerator()
    dataType = rsgisUtils.getRSGISLibDataTypeFromImg(inputImage)
    baseName = os.path.splitext(os.path.basename(inputImage))[0]+"_"+uidStr
    imgTilesDIR = os.path.join(tmpDIR, "imgtiles_"+uidStr)
    tilesClumpsDIR = os.path.join(tmpDIR, "imgclumpstiles_"+uidStr)
    tilesImgBase = os.path.join(imgTilesDIR, baseName)
    initMergedClumps = os.path.join(tmpDIR, "MergedInitClumps_"+uidStr+".kea")
    if not os.path.exists(imgTilesDIR):
        os.makedirs(imgTilesDIR)
    if not os.path.exists(tilesClumpsDIR):
        os.makedirs(tilesClumpsDIR)
    
    imageutils.createTiles(inputImage, tilesImgBase, int(width), int(height), 0, False, 'KEA', dataType, 'kea')
    imageTiles = glob.glob(tilesImgBase+"*")
    
    clumpImgsVals = []
    for tile in imageTiles:
        tilBaseName = os.path.splitext(os.path.basename(tile))[0]
        clumpedTile = os.path.join(tilesClumpsDIR, tilBaseName+'_clumps.kea')
        clumpImgsVals.append([tile, refImg, clumpedTile])
    
    with Pool(nCores) as p:
        p.map(unionClumpImgFunc, clumpImgsVals)
    
    clumpTiles = glob.glob(os.path.join(tilesClumpsDIR, '*_clumps.kea'))
    print("Create Blank Image")
    imageutils.createCopyImage(inputImage, initMergedClumps, 1, 0, 'KEA', rsgislib.TYPE_32UINT)
    print("Merge Tiles into Blank Image")
    segmentation.mergeClumpImages(clumpTiles, initMergedClumps, True)
    print("Merge Tile Boundaries")
    segmentation.mergeEquivClumps(initMergedClumps, clumpsImage, gdalformat, ['ClumpVal_1', 'ClumpVal_2'])
    
    shutil.rmtree(imgTilesDIR)
    shutil.rmtree(tilesClumpsDIR)
    os.remove(initMergedClumps)
    if createdTmp:
        shutil.rmtree(tmpDIR)






#!/usr/bin/env python


############################################################################
#  ratutils.py
#
#  Copyright 2013 RSGISLib.
#
#  RSGISLib: 'The remote sensing and GIS Software Library'
#
#  RSGISLib is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  RSGISLib is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with RSGISLib.  If not, see <http://www.gnu.org/licenses/>.
#
#
# Purpose:  Provide a set of utilities which combine commands to create
#           useful extra functionality and make it more easily available
#           to be reused.
#
# Author: Dan Clewley
# Email: daniel.clewley@gmail.com
# Date: 16/11/2013
# Version: 1.1
#
# History:
# Version 1.0 - Created.
# Version 1.1 - Update to be included into RSGISLib python modules tree
#               (By Pete Bunting).
#
############################################################################

import sys
import math
import os
import os.path
import shutil
import glob
from multiprocessing import Pool
import multiprocessing

from enum import Enum
import rsgislib
from rsgislib import rastergis
from rsgislib import vectorutils
from rsgislib import imageutils
from rsgislib import segmentation

haveGDALPy = True
try:
    import osgeo.gdal as gdal
except ImportError as gdalErr:
    haveGDALPy = False
    
haveGDALOGRPy = True
try:
    from osgeo import ogr
    from osgeo import osr
except ImportError as gdalogrErr:
    haveGDALOGRPy = False

haveMatPlotLib = True
try:
    import matplotlib.pyplot as plt
    import matplotlib.colors as mClrs
except ImportError as pltErr:
    haveMatPlotLib = False   
    
haveNumpy = True
try:
    import numpy
except ImportError as numErr:
    haveNumpy = False

haveHDF5 = True
try:
    import h5py
except ImportError as h5Err:
    haveHDF5 = False
    
haveScipyStats = True
try:
    import scipy.stats
except ImportError as scipystatsErr:
    haveScipyStats = False
    
haveRIOSRat = True
try:
    from rios import rat
    from rios import ratapplier
except ImportError as riosRatErr:
    haveRIOSRat = False

haveSKLearnPCA = True
try:
    from sklearn.decomposition import PCA
except ImportError as sklearnPCAErr:
    haveSKLearnPCA = False


class RSGISRATThresMeasure(Enum):
    kurtosis = 1
    skewness = 2
    combined = 3
    auto = 4
    
class RSGISRATThresDirection(Enum):
    lower = 1
    upper = 2
    lowerupper = 3


def populateImageStats(inputImage, clumpsFile, outascii=None, threshold=0.0, calcMin=False, calcMax=False, calcSum=False, calcMean=False, calcStDev=False, calcMedian=False, calcCount=False, calcArea=False, calcLength=False, calcWidth=False, calcLengthWidth=False):
    """ 
Attribute RAT with statistics from from all bands in an input image.

Where:

:param inputImage: input image to calculate statistics from, if band names are avaialble these will be used for attribute names in the output RAT.
:param clumpsFile: input clumps file, statistics are added to RAT.
:param threshold: float, values below this are ignored (default=0)
:param outascii: string providing output CSV file (optional).
:param calcMin: Calculate minimum
:param calcMax: Calculate maximum
:param calcSum: Calculate sum
:param calcMean: Calculate mean
:param calcStDev: Calculate standard deviation

Example::

    from rsgislib.rastergis import ratutils
    inputImage = 'jers1palsar_stack.kea'
    clumpsFile = 'jers1palsar_stack_clumps_elim_final.kea'
    ratutils.populateImageStats(inputImage, clumpsFile, calcMean=True)

    """
    # Check gdal is available
    if not haveGDALPy:
        raise Exception("The GDAL python bindings required for this function could not be imported\n\t" + gdalErr)
    
    # Open image
    dataset = gdal.Open(inputImage, gdal.GA_ReadOnly)
    
    # Set up list to hold statistics to calculate
    stats2Calc = list()
    
    # Loop through number of bands in image
    nBands = dataset.RasterCount

    # Set up array to hold all column names (used when exporting to ASCII)
    outFieldsList = []
    
    for i in range(nBands):
        bandName = dataset.GetRasterBand(i+1).GetDescription()
        # If band name is not set set to bandN
        if bandName == '':
            bandName = 'Band' + str(i+1)

        # Initialise stats to calculate at None
        minName = None
        maxName = None
        sumName = None
        meanName = None
        stDevName = None
        medianName = None
        countName = None

        if calcMin:
            minName = bandName + 'Min'
            outFieldsList.append(minName)
        if calcMax:
            maxName = bandName + 'Max'
            outFieldsList.append(maxName)
        if calcSum:
            sumName = bandName + 'Sum'
            outFieldsList.append(sumName)
        if calcMean:
            meanName = bandName + 'Avg'
            outFieldsList.append(meanName)
        if calcStDev:
            stDevName = bandName + 'Std'
            outFieldsList.append(stDevName)
        if calcMedian:
            raise Exception('Median is not currently supported.')
            medianName = bandName + 'Med'
            outFieldsList.append(medianName)
        if calcCount:
            raise Exception('Count is not currently supported.')
            countName = bandName + 'Pix'
            outFieldsList.append(countName)

        stats2Calc.append(rastergis.BandAttStats(band=i+1, 
                    minField=minName, maxField=maxName, 
                    sumField=sumName, stdDevField=stDevName, 
                    meanField=meanName))
    
    # Calc stats
    print('''Calculating statistics for %i Bands'''%(nBands))
    t = rsgislib.RSGISTime()
    t.start(True)
    rastergis.populateRATWithStats(inputImage, clumpsFile, stats2Calc)
    t.end()

    # Calculate shapes, if required
    if calcArea or calcLength or calcWidth or calcLengthWidth:
        raise Exception('Shape features are not currently supported.')
        print("\nCalculating shape indices")
        shapes = list()
        if calcArea:
            shapes.append(rastergis.ShapeIndex(colName="Area", idx=rsgislib.SHAPE_SHAPEAREA))
            outFieldsList.append("Area")
        if calcLength:
            shapes.append(rastergis.ShapeIndex(colName="Length", idx=rsgislib.SHAPE_LENGTH))
            outFieldsList.append("Length")
        if calcWidth:
            shapes.append(rastergis.ShapeIndex(colName="Width", idx=rsgislib.SHAPE_WIDTH))
            outFieldsList.append("Width")
        if calcLengthWidth:
            shapes.append(rastergis.ShapeIndex(colName="LengthWidthRatio", idx=rsgislib.SHAPE_LENGTHWIDTH))
            outFieldsList.append("LengthWidthRatio")

        t.start(True)
        rastergis.calcShapeIndices(clumpsFile, shapes)
        t.end()
    
    # Export to ASCII if required
    if outascii is not None:
        print("\nExporting as ASCII")
        t.start(True)
        rastergis.export2Ascii(clumpsFile, outascii, outFieldsList)
        t.end()


def calcPlotGaussianHistoModel(clumpsFile, outGausH5File, outHistH5File, outPlotFile, varCol, binWidth, classColumn, classVal, plotTitle):
    """ Extracts a column from the RAT, masking by a class calculating the histogram and 
        fitting a Gaussian mixture model to the histogram. Outputs include a plot and HDF5
        files of the histogram and gaussian parameters.

Where:

:param clumpsFile: input clumps file with populated RAT.
:param outGausH5File: the output HDF5 file for the Gaussian Mixture Model
:param outHistH5File: the output HDF5 file for the histogram.
:param outPlotFile: the output PDF file for the plot
:param varCol: Column within the RAT for the variable to be used for the histogram
:param binWidth: Bin width for the histogram
:param classColumn: Column where the classes are specified
:param classVal: Class used to mask the input variable
:param plotTitle: title for the plot

Example::

    from rsgislib.rastergis import ratutils
    
    clumpsFile = "FrenchGuiana_10_ALL_sl_HH_lee_UTM_mosaic_dB_segs.kea"
    outGausH5File = "gaufit.h5"
    outHistH5File = "histfile.h5"
    outPlotFile = "Plot.pdf"
    varCol = "HVdB"
    binWidth = 0.1
    classColumn = "Classes"
    classVal = "Mangrove"
    plotTitle = "HV dB Backscater from Mangroves; French Guiana"
    
    ratutils.calcPlotGaussianHistoModel(clumpsFile, outGausH5File, outHistH5File, outPlotFile, varCol, binWidth, classColumn, classVal, plotTitle)

    """
    # Check numpy is available
    if not haveNumpy:
        raise Exception("The numpy module is required for this function could not be imported\n\t" + numErr)
    # Check gdal is available
    if not haveGDALPy:
        raise Exception("The GDAL python bindings required for this function could not be imported\n\t" + gdalErr)
    # Check matplotlib is available
    if not haveMatPlotLib:
        raise Exception("The matplotlib module is required for this function could not be imported\n\t" + pltErr)       
    # Check hdf5 is available
    if not haveHDF5:
        raise Exception("The hdf5 module is required for this function could not be imported\n\t" + h5Err)
        
    # Calculate histogram and fit Gaussian Mixture Model
    rastergis.fitHistGausianMixtureModel(clumps=clumpsFile, outH5File=outGausH5File, outHistFile=outHistH5File, varCol=varCol, binWidth=binWidth, classColumn=classColumn, classVal=classVal)
    
    
    if not h5py.is_hdf5(outGausH5File):
        raise Exception(outGausH5File + " is not a HDF5 file.")
        
    if not h5py.is_hdf5(outHistH5File):
        raise Exception(outHistH5File + " is not a HDF5 file.")

    gausFile = h5py.File(outGausH5File,'r')
    gausParams = gausFile['/DATA/DATA']
    
    
    histFile = h5py.File(outHistH5File,'r')
    histData = histFile['/DATA/DATA']
    
    xVals = []
    xValsHist = []
    histBins = []
    
    for histBin in histData:
        xValsHist.append(histBin[0]-(binWidth/2))
        xVals.append(histBin[0])
        histBins.append(histBin[1])
            
    gAmpVals = []
    gFWHMVals = []
    gOffVals = []
    gNoiseVals = []
    noiseVal = 0.0
    
    for gausParam in gausParams:
        gOffVals.append(gausParam[0])
        gAmpVals.append(gausParam[1])
        gFWHMVals.append(gausParam[2])
        noiseVal = gausParam[3]
    
    fig, ax = plt.subplots()
    histBars = ax.bar(xValsHist, histBins, width=binWidth, color='#A7A7A7', edgecolor='#A7A7A7')
    
    predVals = numpy.zeros(len(xVals))
    for i in range(len(xVals)):
        gNoiseVals.append(noiseVal)
        for j in range(len(gOffVals)):
            predVals[i] = predVals[i] + (gAmpVals[j] * math.exp((-1.0)*(pow(xVals[i] - gOffVals[j], 2)/(2.0 * pow(gFWHMVals[j], 2)))))
        predVals[i] = predVals[i] + noiseVal
    
    ax.plot(xVals, predVals, color='red')
    ax.plot(xVals, gNoiseVals, color='blue', linestyle='dashed')
    
    ax.set_ylabel('Freq.')
    ax.set_title(plotTitle)
    plt.savefig(outPlotFile, format='PDF')
    
    gausFile.close()
    histFile.close()
    

def findChangeClumpsHistSkewKurtTest(inputClumps, inClassCol, classOfInterest, changeVarCol, outChangeFeatCol, noDataVals=[], thresMeasure=RSGISRATThresMeasure.auto, exportPlot=None, showAllThreshPlot=False):
    """
This function identifies potential change features from both sides of the histogram 
by slicing the histogram and finding an optimal skewness and kurtosis.

Where:

:param inputClumps: input clumps file.
:param inClassCol: The column specifiying the classes, one of which change will be found.
:param classOfInterest: The class (as defined in inClassCol) on which changed is being found.
:param changeVarCol: Variable(s) to be used to find change. Expecting column name. Needs to be numeric. If a list of column names is provided then they are combined using PCA and the first PC is used for the change process.
:param outChangeFeatCol: the output column. Regions lower than lower threshold have value 1. Regions higher than upper threshold have value 2. No change had threshold 0.
:param noDataVals: list of no data values to be ignored.
:param thresMeasure: needs to be of type RSGISRATThresMeasure (default is auto)
:param exportPlot: file name for exporting a histogram plot with thresholds annotated. No plot is create if None is passed (default is none).
:param showAllThreshPlot: option when plotting to put all the thresholds on to the plot rather than just the one being used.

:return: list of lower [0] and upper [1] thresholds used to define the no change region.

    """
    # Check numpy is available
    if not haveNumpy:
        raise Exception("The numpy module is required for this function could not be imported\n\t" + numErr)
    # Check gdal is available
    if not haveGDALPy:
        raise Exception("The GDAL python bindings are required for this function could not be imported\n\t" + gdalErr)
    # Check rios rat is available
    if not haveRIOSRat:
        raise Exception("The RIOS rat tools are required for this function could not be imported\n\t" + riosRatErr)
    # Check scipy stats is available
    if not haveScipyStats:
        raise Exception("The scipy stats is required for this function could not be imported\n\t" + scipystatsErr)
    if not exportPlot == None:
        # Check matplotlib is available
        if not haveMatPlotLib:
            raise Exception("The matplotlib module is required for this function could not be imported\n\t" + pltErr)
    if type(changeVarCol) is list:
        if not haveSKLearnPCA:
            raise Exception("The scikit learn library PCA module is required when a list of column variables is given\n\t" + sklearnPCAErr)

    ## Open the image file...
    ratDataset = gdal.Open(inputClumps, gdal.GA_Update)

    ## Read in columns
    classVals = rat.readColumn(ratDataset, inClassCol)
    outChangeFeats = numpy.zeros_like(classVals)
    ID = numpy.arange(classVals.shape[0])
    
    vals = None
    if type(changeVarCol) is list:
        numVars = len(changeVarCol)
        numRows = classVals.shape[0]
        varVals = numpy.zeros((numVars,numRows), dtype=numpy.float)
        i = 0
        for varCol in changeVarCol:
            colVals = rat.readColumn(ratDataset, varCol)
            varVals[i] = colVals
            i = i + 1
        varVals = varVals.transpose()
        
        ID = ID[classVals == classOfInterest]
        varVals = varVals[(classVals == classOfInterest)]
        
        ID = ID[numpy.isfinite(varVals).all(axis=1)]
        varVals = varVals[numpy.isfinite(varVals).all(axis=1)]
        
        for noDataVal in noDataVals:
            ID = ID[(varVals != noDataVal).all(axis=1)]
            varVals = varVals[(varVals != noDataVal).all(axis=1)]
        
        pca = PCA(n_components=1)
        fittedPCA = pca.fit(varVals)
    
        vals = fittedPCA.transform(varVals)[:,0]
    else:
        vals = rat.readColumn(ratDataset, changeVarCol)        
    
        ID = ID[classVals == classOfInterest]
        if ID.shape[0] == 0:
            rat.writeColumn(ratDataset, outChangeFeatCol, outChangeFeats)
            return
        vals = vals[classVals == classOfInterest]
        
        ID = ID[numpy.isfinite(vals)]
        vals = vals[numpy.isfinite(vals)]
        
        for noDataVal in noDataVals:
            ID = ID[vals != noDataVal]
            vals = vals[vals != noDataVal]
    
    n = vals.shape[0]
    lq = numpy.percentile(vals, 25)
    uq = numpy.percentile(vals, 75)
    iqr = uq - lq
    binSize = 2 * iqr * n**(-1/3)
    print("Bin Size = ", binSize)
    numBins = int((numpy.max(vals) - numpy.min(vals))/binSize)+2
    print("num of bins = ", numBins)
    
    hist, bin_edges = numpy.histogram(vals, bins=numBins)
    
    print(hist.shape)
    print(bin_edges.shape)
        
    print("LQ = ", lq)
    print("UQ = ", uq)
    
    lqNumBins = int((lq - bin_edges[0])/binSize)+1
    uqNumBins = int((bin_edges[-1]-uq)/binSize)+1
    
    print("lqNumBins = ", lqNumBins)
    print("uqNumBins = ", uqNumBins)
    
    kurtosisVals = numpy.zeros((lqNumBins,uqNumBins), dtype=numpy.float)
    skewnessVals = numpy.zeros((lqNumBins,uqNumBins), dtype=numpy.float)
    lowBins = numpy.zeros((lqNumBins,uqNumBins), dtype=numpy.int)
    upBins = numpy.zeros((lqNumBins,uqNumBins), dtype=numpy.int)
    
    for lowBin in range(lqNumBins):
        for upBin in range(uqNumBins):
            #print("Bin [" + str(lowBin) + ", " + str(numBins-upBin) + "]")
            histTmp = hist[lowBin:(numBins-upBin)]
            #print(histTmp)
            #print(histTmp.shape)
            lowBins[lowBin,upBin] = lowBin
            upBins[lowBin,upBin] = numBins-upBin
            
            kurtosisVals[lowBin,upBin] = scipy.stats.kurtosis(histTmp)
            skewnessVals[lowBin,upBin] = scipy.stats.skew(histTmp)
            
    
    #print(kurtosisVals)
    #print(skewnessVals)
    kurtosisValsAbs = numpy.absolute(kurtosisVals)
    skewnessValsAbs = numpy.absolute(skewnessVals)
    #print("Kurtosis Range: [" + str(numpy.min(kurtosisValsAbs)) + ", " + str(numpy.max(kurtosisValsAbs)) + "]") 
    #print("Skewness Range: [" + str(numpy.min(skewnessValsAbs)) + ", " + str(numpy.max(skewnessValsAbs)) + "]") 
    kurtosisValsNorm = (kurtosisValsAbs-numpy.min(kurtosisValsAbs)) / (numpy.max(kurtosisValsAbs)-numpy.min(kurtosisValsAbs))
    skewnessValsNorm = (skewnessValsAbs-numpy.min(skewnessValsAbs)) / (numpy.max(skewnessValsAbs)-numpy.min(skewnessValsAbs))
    
    #print("Kurtosis Norm Range: [" + str(numpy.min(kurtosisValsNorm)) + ", " + str(numpy.max(kurtosisValsNorm)) + "]") 
    #print("Skewness Norm Range: [" + str(numpy.min(skewnessValsNorm)) + ", " + str(numpy.max(skewnessValsNorm)) + "]") 
    
    combined = kurtosisValsNorm + skewnessValsNorm
    #combined = kurtosisValsAbs + skewnessValsAbs
    #print(combined)
    
    minKurt = numpy.unravel_index(numpy.argmin(kurtosisValsAbs, axis=None), kurtosisValsAbs.shape)
    minSkew = numpy.unravel_index(numpy.argmin(skewnessValsAbs, axis=None), skewnessValsAbs.shape)
    minComb = numpy.unravel_index(numpy.argmin(combined, axis=None), combined.shape)
    
    print("Kurtosis bin indexes: ", minKurt)
    print("Skewness bin indexes: ", minSkew)
    print("Combined bin indexes: ", minComb)
    
    lowBinKurt = minKurt[0]
    lowerThresKurt = bin_edges[lowBinKurt] + (binSize/2)
    upBinKurt = numBins-minKurt[1]
    upperThresKurt = bin_edges[upBinKurt] + (binSize/2)
    print("No Change Data Range (Kurtosis): [" + str(lowerThresKurt) + "," + str(upperThresKurt) + "]")
    
    lowBinSkew = minSkew[0]
    lowerThresSkew = bin_edges[lowBinSkew] + (binSize/2)
    upBinSkew = numBins-minSkew[1]
    upperThresSkew = bin_edges[upBinSkew] + (binSize/2)
    print("No Change Data Range (Skewness): [" + str(lowerThresSkew) + "," + str(upperThresSkew) + "]")
        
    lowBinComb = minComb[0]
    lowerThresComb = bin_edges[lowBinComb] + (binSize/2)
    upBinComb = numBins-minComb[1]
    upperThresComb = bin_edges[upBinComb] + (binSize/2)
    print("No Change Data Range (Combined): [" + str(lowerThresComb) + "," + str(upperThresComb) + "]")
    
    lowerThres = 0.0
    upperThres = 0.0
    if thresMeasure == RSGISRATThresMeasure.kurtosis:
        lowerThres = lowerThresKurt
        upperThres = upperThresKurt
    elif thresMeasure == RSGISRATThresMeasure.skewness:
        lowerThres = lowerThresSkew
        upperThres = upperThresSkew
    elif thresMeasure == RSGISRATThresMeasure.combined:
        lowerThres = lowerThresComb
        upperThres = upperThresComb
    elif thresMeasure == RSGISRATThresMeasure.auto:
        if (abs(lowerThresKurt-lowerThresSkew) > (uq-lq)) or (abs(upperThresKurt-upperThresSkew) > (uq-lq)):
            lowerThres = lowerThresSkew
            upperThres = upperThresSkew
        else:
            lowerThres = lowerThresComb
            upperThres = upperThresComb
        print("No Change Data Range (auto): [" + str(lowerThres) + "," + str(upperThres) + "]")
    else:
        raise Exception("Don't understand metric for threshold provided must be of type ThresMeasure")
    

    if not exportPlot == None:
        center = (bin_edges[:-1] + bin_edges[1:]) / 2
        plt.bar(center, hist, align='center', width=binSize)
        if showAllThreshPlot:
            plt.vlines(lowerThresKurt, 0, numpy.max(hist), color='y', linewidth=1, label='Kurtosis Lower')
            plt.vlines(upperThresKurt, 0, numpy.max(hist), color='y', linewidth=1, label='Kurtosis Upper')
            plt.vlines(lowerThresSkew, 0, numpy.max(hist), color='r', linewidth=1, label='Skewness Lower')
            plt.vlines(upperThresSkew, 0, numpy.max(hist), color='r', linewidth=1, label='Skewness Upper')
            plt.vlines(lowerThresComb, 0, numpy.max(hist), color='g', linewidth=1, label='Combined Lower')
            plt.vlines(upperThresComb, 0, numpy.max(hist), color='g', linewidth=1, label='Combined Upper')
        else:
            plt.vlines(lowerThres, 0, numpy.max(hist), color='r', linewidth=1, label='Lower Threshold')
            plt.vlines(upperThres, 0, numpy.max(hist), color='r', linewidth=1, label='Upper Threshold')
        plt.grid(True)
        plt.legend(loc=0)
        plt.savefig(exportPlot)
        plt.close()    
    
    ## Apply to RAT...
    changeFeats = numpy.where(vals < lowerThres, 1, 0)
    changeFeats = numpy.where(vals > upperThres, 2, changeFeats)
    
    outChangeFeats[ID] = changeFeats
    rat.writeColumn(ratDataset, outChangeFeatCol, outChangeFeats)
    
    ratDataset = None
    return [lowerThres, upperThres]


def findChangeClumpsHistSkewKurtTestLower(inputClumps, inClassCol, classOfInterest, changeVarCol, outChangeFeatCol, noDataVals=[], thresMeasure=RSGISRATThresMeasure.auto, exportPlot=None, showAllThreshPlot=False):
    """
This function identifies potential change features from just the lower (left) side of the histogram 
by slicing the histogram and finding an optimal skewness and kurtosis.

Where:

:param inputClumps: input clumps file.
:param inClassCol: The column specifiying the classes, one of which change will be found.
:param classOfInterest: The class (as defined in inClassCol) on which changed is being found.
:param changeVarCol: changeVarCol - Variable(s) to be used to find change. Expecting column name. Needs to be numeric. If a list of column names is provided then they are combined using PCA and the first PC is used for the change process.
:param outChangeFeatCol: the output column. Regions lower than lower threshold have value 1. Regions higher than upper threshold have value 2. No change had threshold 0.
:param noDataVals: list of no data values to be ignored.
:param thresMeasure: needs to be of type RSGISRATThresMeasure (default is auto)
:param exportPlot: file name for exporting a histogram plot with thresholds annotated. No plot is create if None is passed (default is none).
:param showAllThreshPlot: option when plotting to put all the thresholds on to the plot rather than just the one being used.

:return: list of lower [0] and upper [1] thresholds used to define the no change region.
    
    """
    # Check numpy is available
    if not haveNumpy:
        raise Exception("The numpy module is required for this function could not be imported\n\t" + numErr)
    # Check gdal is available
    if not haveGDALPy:
        raise Exception("The GDAL python bindings are required for this function could not be imported\n\t" + gdalErr)
    # Check rios rat is available
    if not haveRIOSRat:
        raise Exception("The RIOS rat tools are required for this function could not be imported\n\t" + riosRatErr)
    # Check scipy stats is available
    if not haveScipyStats:
        raise Exception("The scipy stats is required for this function could not be imported\n\t" + scipystatsErr)
    if not exportPlot == None:
        # Check matplotlib is available
        if not haveMatPlotLib:
            raise Exception("The matplotlib module is required for this function could not be imported\n\t" + pltErr)
    if type(changeVarCol) is list:
        if not haveSKLearnPCA:
            raise Exception("The scikit learn library PCA module is required when a list of column variables is given\n\t" + sklearnPCAErr)
    ## Open the image file...
    ratDataset = gdal.Open(inputClumps, gdal.GA_Update)

    ## Read in columns
    classVals = rat.readColumn(ratDataset, inClassCol)
    outChangeFeats = numpy.zeros_like(classVals)
    ID = numpy.arange(classVals.shape[0])
    
    vals = None
    if type(changeVarCol) is list:
        numVars = len(changeVarCol)
        numRows = classVals.shape[0]
        varVals = numpy.zeros((numVars,numRows), dtype=numpy.float)
        i = 0
        for varCol in changeVarCol:
            colVals = rat.readColumn(ratDataset, varCol)
            varVals[i] = colVals
            i = i + 1
        varVals = varVals.transpose()
        
        ID = ID[classVals == classOfInterest]
        varVals = varVals[(classVals == classOfInterest)]
        
        ID = ID[numpy.isfinite(varVals).all(axis=1)]
        varVals = varVals[numpy.isfinite(varVals).all(axis=1)]
        
        for noDataVal in noDataVals:
            ID = ID[(varVals != noDataVal).all(axis=1)]
            varVals = varVals[(varVals != noDataVal).all(axis=1)]
        
        pca = PCA(n_components=1)
        fittedPCA = pca.fit(varVals)
    
        vals = fittedPCA.transform(varVals)[:,0]
    else:
        vals = rat.readColumn(ratDataset, changeVarCol)        
    
        ID = ID[classVals == classOfInterest]
        if ID.shape[0] == 0:
            rat.writeColumn(ratDataset, outChangeFeatCol, outChangeFeats)
            return
        vals = vals[classVals == classOfInterest]
        
        ID = ID[numpy.isfinite(vals)]
        vals = vals[numpy.isfinite(vals)]
        
        for noDataVal in noDataVals:
            ID = ID[vals != noDataVal]
            vals = vals[vals != noDataVal]

    n = vals.shape[0]
    lq = numpy.percentile(vals, 25)
    uq = numpy.percentile(vals, 75)
    iqr = uq - lq
    binSize = 2 * iqr * n**(-1/3)
    print("Bin Size = ", binSize)
    numBins = int((numpy.max(vals) - numpy.min(vals))/binSize)+2
    print("num of bins = ", numBins)
    
    hist, bin_edges = numpy.histogram(vals, bins=numBins)
    
    print(hist.shape)
    print(bin_edges.shape)
        
    print("LQ = ", lq)
    print("UQ = ", uq)
    
    lqNumBins = int((lq - bin_edges[0])/binSize)+1
    
    print("lqNumBins = ", lqNumBins)
    
    kurtosisVals = numpy.zeros((lqNumBins), dtype=numpy.float)
    skewnessVals = numpy.zeros((lqNumBins), dtype=numpy.float)
    lowBins = numpy.zeros((lqNumBins), dtype=numpy.int)
    
    for lowBin in range(lqNumBins):
        histTmp = hist[lowBin:-1]
        lowBins[lowBin] = lowBin
        
        kurtosisVals[lowBin] = scipy.stats.kurtosis(histTmp)
        skewnessVals[lowBin] = scipy.stats.skew(histTmp)

    kurtosisValsAbs = numpy.absolute(kurtosisVals)
    skewnessValsAbs = numpy.absolute(skewnessVals)
    print("Kurtosis Range: [" + str(numpy.min(kurtosisValsAbs)) + ", " + str(numpy.max(kurtosisValsAbs)) + "]") 
    print("Skewness Range: [" + str(numpy.min(skewnessValsAbs)) + ", " + str(numpy.max(skewnessValsAbs)) + "]") 
    kurtosisValsNorm = (kurtosisValsAbs-numpy.min(kurtosisValsAbs)) / (numpy.max(kurtosisValsAbs)-numpy.min(kurtosisValsAbs))
    skewnessValsNorm = (skewnessValsAbs-numpy.min(skewnessValsAbs)) / (numpy.max(skewnessValsAbs)-numpy.min(skewnessValsAbs))

    combined = kurtosisValsNorm + skewnessValsNorm

    minKurt = numpy.argmin(kurtosisValsAbs)
    minSkew = numpy.argmin(skewnessValsAbs)
    minComb = numpy.argmin(combined)
    
    print("Kurtosis bin index: ", minKurt)
    print("Skewness bin index: ", minSkew)
    print("Combined bin index: ", minComb)
    
    lowBinKurt = minKurt
    lowerThresKurt = bin_edges[lowBinKurt] + (binSize/2)
    print("Lower Threshold (Kurtosis): " + str(lowerThresKurt))
    
    lowBinSkew = minSkew
    lowerThresSkew = bin_edges[lowBinSkew] + (binSize/2)
    print("Lower Threshold (Skewness): " + str(lowerThresSkew))
        
    lowBinComb = minComb
    lowerThresComb = bin_edges[lowBinComb] + (binSize/2)
    print("Lower Threshold (Combined): " + str(lowerThresComb))
    
    lowerThres = 0.0
    upperThres = numpy.max(vals)
    if thresMeasure == RSGISRATThresMeasure.kurtosis:
        lowerThres = lowerThresKurt
    elif thresMeasure == RSGISRATThresMeasure.skewness:
        lowerThres = lowerThresSkew
    elif thresMeasure == RSGISRATThresMeasure.combined:
        lowerThres = lowerThresComb
    elif thresMeasure == RSGISRATThresMeasure.auto:
        if abs(lowerThresKurt-lowerThresSkew) > (uq-lq):
            lowerThres = lowerThresSkew
        else:
            lowerThres = lowerThresComb        
        print("Lower Threshold (auto): " + str(lowerThres))
    else:
        raise Exception("Don't understand metric for threshold provided must be of type ThresMeasure")
    
    if not exportPlot == None:
        center = (bin_edges[:-1] + bin_edges[1:]) / 2
        plt.bar(center, hist, align='center', width=binSize)
        if showAllThreshPlot:
            plt.vlines(lowerThresKurt, 0, numpy.max(hist), color='y', linewidth=1, label='Kurtosis Lower')
            plt.vlines(lowerThresSkew, 0, numpy.max(hist), color='r', linewidth=1, label='Skewness Lower')
            plt.vlines(lowerThresComb, 0, numpy.max(hist), color='g', linewidth=1, label='Combined Lower')
        else:
            plt.vlines(lowerThres, 0, numpy.max(hist), color='r', linewidth=1, label='Lower Threshold')
        plt.grid(True)
        plt.legend(loc=0)
        plt.savefig(exportPlot)
        plt.close()
    
    ## Apply to RAT...
    changeFeats = numpy.where(vals < lowerThres, 1, 0)
    
    outChangeFeats[ID] = changeFeats
    rat.writeColumn(ratDataset, outChangeFeatCol, outChangeFeats)
    
    ratDataset = None
    return [lowerThres, upperThres]


def findChangeClumpsHistSkewKurtTestUpper(inputClumps, inClassCol, classOfInterest, changeVarCol, outChangeFeatCol, noDataVals=[], thresMeasure=RSGISRATThresMeasure.auto, exportPlot=None, showAllThreshPlot=False):
    """
This function identifies potential change features from just the upper (right) side of the histogram 
by slicing the histogram and finding an optimal skewness and kurtosis.

Where:

:param inputClumps: input clumps file.
:param inClassCol: The column specifiying the classes, one of which change will be found.
:param classOfInterest: The class (as defined in inClassCol) on which changed is being found.
:param changeVarCol: changeVarCol - Variable(s) to be used to find change. Expecting column name. Needs to be numeric. If a list of column names is provided then they are combined using PCA and the first PC is used for the change process.
:param outChangeFeatCol: the output column. Regions lower than lower threshold have value 1. Regions higher than upper threshold have value 2. No change had threshold 0.
:param noDataVals: list of no data values to be ignored.
:param thresMeasure: needs to be of type RSGISRATThresMeasure (default is auto)
:param exportPlot: file name for exporting a histogram plot with thresholds annotated. No plot is create if None is passed (default is none).
:param showAllThreshPlot: option when plotting to put all the thresholds on to the plot rather than just the one being used.

:return: list of lower [0] and upper [1] thresholds used to define the no change region.

    """
    # Check numpy is available
    if not haveNumpy:
        raise Exception("The numpy module is required for this function could not be imported\n\t" + numErr)
    # Check gdal is available
    if not haveGDALPy:
        raise Exception("The GDAL python bindings are required for this function could not be imported\n\t" + gdalErr)
    # Check rios rat is available
    if not haveRIOSRat:
        raise Exception("The RIOS rat tools are required for this function could not be imported\n\t" + riosRatErr)
    # Check scipy stats is available
    if not haveScipyStats:
        raise Exception("The scipy stats is required for this function could not be imported\n\t" + scipystatsErr)
    if not exportPlot == None:
        # Check matplotlib is available
        if not haveMatPlotLib:
            raise Exception("The matplotlib module is required for this function could not be imported\n\t" + pltErr)
    if type(changeVarCol) is list:
        if not haveSKLearnPCA:
            raise Exception("The scikit learn library PCA module is required when a list of column variables is given\n\t" + sklearnPCAErr)
    ## Open the image file...
    ratDataset = gdal.Open(inputClumps, gdal.GA_Update)

    ## Read in columns
    classVals = rat.readColumn(ratDataset, inClassCol)
    outChangeFeats = numpy.zeros_like(classVals)
    ID = numpy.arange(classVals.shape[0])
    
    vals = None
    if type(changeVarCol) is list:
        numVars = len(changeVarCol)
        numRows = classVals.shape[0]
        varVals = numpy.zeros((numVars,numRows), dtype=numpy.float)
        i = 0
        for varCol in changeVarCol:
            colVals = rat.readColumn(ratDataset, varCol)
            varVals[i] = colVals
            i = i + 1
        varVals = varVals.transpose()
        
        ID = ID[classVals == classOfInterest]
        varVals = varVals[(classVals == classOfInterest)]
        
        ID = ID[numpy.isfinite(varVals).all(axis=1)]
        varVals = varVals[numpy.isfinite(varVals).all(axis=1)]
        
        for noDataVal in noDataVals:
            ID = ID[(varVals != noDataVal).all(axis=1)]
            varVals = varVals[(varVals != noDataVal).all(axis=1)]
        
        pca = PCA(n_components=1)
        fittedPCA = pca.fit(varVals)
    
        vals = fittedPCA.transform(varVals)[:,0]
    else:
        vals = rat.readColumn(ratDataset, changeVarCol)        
    
        ID = ID[classVals == classOfInterest]
        if ID.shape[0] == 0:
            rat.writeColumn(ratDataset, outChangeFeatCol, outChangeFeats)
            return
        vals = vals[classVals == classOfInterest]
        
        ID = ID[numpy.isfinite(vals)]
        vals = vals[numpy.isfinite(vals)]
        
        for noDataVal in noDataVals:
            ID = ID[vals != noDataVal]
            vals = vals[vals != noDataVal]
    
    n = vals.shape[0]
    lq = numpy.percentile(vals, 25)
    uq = numpy.percentile(vals, 75)
    iqr = uq - lq
    binSize = 2 * iqr * n**(-1/3)
    print("Bin Size = ", binSize)
    numBins = int((numpy.max(vals) - numpy.min(vals))/binSize)+2
    print("num of bins = ", numBins)
    
    hist, bin_edges = numpy.histogram(vals, bins=numBins)
    
    print(hist.shape)
    print(bin_edges.shape)
        
    print("LQ = ", lq)
    print("UQ = ", uq)
    
    uqNumBins = int((bin_edges[-1]-uq)/binSize)+1
    
    print("uqNumBins = ", uqNumBins)
    
    kurtosisVals = numpy.zeros((uqNumBins), dtype=numpy.float)
    skewnessVals = numpy.zeros((uqNumBins), dtype=numpy.float)
    upBins = numpy.zeros((uqNumBins), dtype=numpy.int)
    
    for upBin in range(uqNumBins):
        histTmp = hist[0:(numBins-upBin)]
        upBins[upBin] = numBins-upBin
        
        kurtosisVals[upBin] = scipy.stats.kurtosis(histTmp)
        skewnessVals[upBin] = scipy.stats.skew(histTmp)

    kurtosisValsAbs = numpy.absolute(kurtosisVals)
    skewnessValsAbs = numpy.absolute(skewnessVals)
    print("Kurtosis Range: [" + str(numpy.min(kurtosisValsAbs)) + ", " + str(numpy.max(kurtosisValsAbs)) + "]") 
    print("Skewness Range: [" + str(numpy.min(skewnessValsAbs)) + ", " + str(numpy.max(skewnessValsAbs)) + "]") 
    kurtosisValsNorm = (kurtosisValsAbs-numpy.min(kurtosisValsAbs)) / (numpy.max(kurtosisValsAbs)-numpy.min(kurtosisValsAbs))
    skewnessValsNorm = (skewnessValsAbs-numpy.min(skewnessValsAbs)) / (numpy.max(skewnessValsAbs)-numpy.min(skewnessValsAbs))

    combined = kurtosisValsNorm + skewnessValsNorm
    
    minKurt = numpy.argmin(kurtosisValsAbs)
    minSkew = numpy.argmin(skewnessValsAbs)
    minComb = numpy.argmin(combined)
    
    print("Kurtosis bin index: ", minKurt)
    print("Skewness bin index: ", minSkew)
    print("Combined bin index: ", minComb)
    
    upBinKurt = numBins-minKurt
    upperThresKurt = bin_edges[upBinKurt] + (binSize/2)
    print("Upper Threshold (Kurtosis): " + str(upperThresKurt))
    
    upBinSkew = numBins-minSkew
    upperThresSkew = bin_edges[upBinSkew] + (binSize/2)
    print("Upper Threshold (Skewness): " + str(upperThresSkew))
        
    upBinComb = numBins-minComb
    upperThresComb = bin_edges[upBinComb] + (binSize/2)
    print("Upper Threshold (Combined): " + str(upperThresComb))
    
    lowerThres = numpy.min(vals)
    upperThres = 0.0
    if thresMeasure == RSGISRATThresMeasure.kurtosis:
        upperThres = upperThresKurt
    elif thresMeasure == RSGISRATThresMeasure.skewness:
        upperThres = upperThresSkew
    elif thresMeasure == RSGISRATThresMeasure.combined:
        upperThres = upperThresComb
    elif thresMeasure == RSGISRATThresMeasure.auto:
        if abs(upperThresKurt-upperThresSkew) > (uq-lq):
            upperThres = upperThresSkew
        else:
            upperThres = upperThresComb
        print("Upper Threshold (auto): " + str(upperThres))
    else:
        raise Exception("Don't understand metric for threshold provided must be of type ThresMeasure")
    
    if not exportPlot == None:
        center = (bin_edges[:-1] + bin_edges[1:]) / 2
        plt.bar(center, hist, align='center', width=binSize)
        if showAllThreshPlot:
            plt.vlines(upperThresKurt, 0, numpy.max(hist), color='y', linewidth=1, label='Kurtosis Upper')
            plt.vlines(upperThresSkew, 0, numpy.max(hist), color='r', linewidth=1, label='Skewness Upper')
            plt.vlines(upperThresComb, 0, numpy.max(hist), color='g', linewidth=1, label='Combined Upper')
        else:
            plt.vlines(upperThres, 0, numpy.max(hist), color='r', linewidth=1, label='Upper Threshold')
        plt.grid(True)
        plt.legend(loc=0)
        plt.savefig(exportPlot)
        plt.close()
        
    ## Apply to RAT...
    changeFeats = numpy.where(vals > upperThres, 1, 0)
    
    outChangeFeats[ID] = changeFeats
    rat.writeColumn(ratDataset, outChangeFeatCol, outChangeFeats)
    
    ratDataset = None
    return [lowerThres, upperThres]


class RSGISRATChangeVarInfo:
    """
    A class to store the change variable information required for some of the change functions.
    """
    def __init__(self, changeVarCol="", outChangeFeatCol="", noDataVals=[], thresMeasure=RSGISRATThresMeasure.auto, thresDirection=RSGISRATThresDirection.lower, exportPlot=None, showAllThreshPlot=False, lowerThreshold=0.0, upperThreshold=0.0):
        self.changeVarCol = changeVarCol
        self.outChangeFeatCol = outChangeFeatCol
        self.noDataVals = noDataVals
        self.thresMeasure = thresMeasure
        self.thresDirection = thresDirection
        self.exportPlot = exportPlot
        self.showAllThreshPlot = showAllThreshPlot
        self.lowerThreshold = lowerThreshold
        self.upperThreshold = upperThreshold


def findChangeClumpsHistSkewKurtTestVoteMultiVars(inputClumps, inClassCol, classOfInterest, outChangeFeatCol, vars=[]):
    """
A function to call one of the findChangeClumpsHistSkewKurtTest functions for multiple 
variables and then combine together by voting to find change features.

Where:

:param inputClumps: input clumps file.
:param inClassCol: The column specifiying the classes, one of which change will be found.
:param classOfInterest: The class (as defined in inClassCol) on which changed is being found.
:param outChangeFeatCol: the output column with the vote scores.
:param vars: a list of RSGISRATChangeVarInfo objects used to specify the variables and function to be called.

    """
    # Check numpy is available
    if not haveNumpy:
        raise Exception("The numpy module is required for this function could not be imported\n\t" + numErr)
    # Check gdal is available
    if not haveGDALPy:
        raise Exception("The GDAL python bindings are required for this function could not be imported\n\t" + gdalErr)
    # Check rios rat is available
    if not haveRIOSRat:
        raise Exception("The RIOS rat tools are required for this function could not be imported\n\t" + riosRatErr)
    if len(vars) == 0:
        raise Exception("Need to provide a list of variables with parameters...")
    for var in vars:
        print(var.changeVarCol)
        if var.thresDirection == RSGISRATThresDirection.lower:
            outThres = findChangeClumpsHistSkewKurtTestLower(inputClumps, inClassCol, classOfInterest, var.changeVarCol, var.outChangeFeatCol, var.noDataVals, var.thresMeasure, var.exportPlot, var.showAllThreshPlot)
            var.lowerThreshold = outThres[0]
            var.upperThreshold = outThres[1]
        elif var.thresDirection == RSGISRATThresDirection.upper:
            outThres = findChangeClumpsHistSkewKurtTestUpper(inputClumps, inClassCol, classOfInterest, var.changeVarCol, var.outChangeFeatCol, var.noDataVals, var.thresMeasure, var.exportPlot, var.showAllThreshPlot)
            var.lowerThreshold = outThres[0]
            var.upperThreshold = outThres[1]
        elif var.thresDirection == RSGISRATThresDirection.lowerupper:
            outThres = findChangeClumpsHistSkewKurtTest(inputClumps, inClassCol, classOfInterest, var.changeVarCol, var.outChangeFeatCol, var.noDataVals, var.thresMeasure, var.exportPlot, var.showAllThreshPlot)
            var.lowerThreshold = outThres[0]
            var.upperThreshold = outThres[1]
        else:
            raise Exception("Direction must be of type RSGISRATThresDirection and only lower, upper and lowerupper are supported")
        
    ratDataset = gdal.Open(inputClumps, gdal.GA_Update)
    classVals = rat.readColumn(ratDataset, inClassCol)
    changeVote = numpy.zeros_like(classVals, dtype=numpy.int)

    for var in vars:
        changeCol = rat.readColumn(ratDataset, var.outChangeFeatCol)
        if var.thresDirection == RSGISRATThresDirection.lowerupper:
            changeCol[changeCol == 2] = 1
        changeVote = changeVote + changeCol
    
    rat.writeColumn(ratDataset, outChangeFeatCol, changeVote)
    ratDataset = None


def findClumpsWithinExistingThresholds(inputClumps, inClassCol, classOfInterest, outFeatsCol, vars=[]):
    """
A function to use the thresholds stored in the RSGISRATChangeVarInfo objects (var) 
and populated from the findChangeClumpsHistSkewKurtTest functions to assess another class
creating a binary column as to whether a feature is within the threshold or now. Where multiple
variables (i.e., len(var) > 1) then variables are combined with an and operation.

Where:

:param inputClumps: input clumps file.
:param inClassCol: The column specifiying the classes, one of which change will be found.
:param classOfInterest: The class (as defined in inClassCol) on which changed is being found.
:param outFeatsCol: the output binary column specifying whether a feature is within the thresholds.
:param vars: a list of RSGISRATChangeVarInfo objects used to specify the variables and function to be called.

    """
    # Check numpy is available
    if not haveNumpy:
        raise Exception("The numpy module is required for this function could not be imported\n\t" + numErr)
    # Check gdal is available
    if not haveGDALPy:
        raise Exception("The GDAL python bindings are required for this function could not be imported\n\t" + gdalErr)
    # Check rios rat is available
    if not haveRIOSRat:
        raise Exception("The RIOS rat tools are required for this function could not be imported\n\t" + riosRatErr)
    if len(vars) == 0:
        raise Exception("Need to provide a list of variables with parameters...")
        
    ## Open the image file...
    ratDataset = gdal.Open(inputClumps, gdal.GA_Update)

    ## Read in columns
    classVals = rat.readColumn(ratDataset, inClassCol)
    outFeats = numpy.zeros_like(classVals)
    
    first = True
    for var in vars:
        print(var.changeVarCol)
        if first:
            varVals = rat.readColumn(ratDataset, var.changeVarCol)
            outFeats = numpy.where( (varVals > var.lowerThreshold) & (varVals < var.upperThreshold) & (classVals == classOfInterest), 1, outFeats)
            first = False
        else:
            varVals = rat.readColumn(ratDataset, var.changeVarCol)
            outFeats = numpy.where( (varVals > var.lowerThreshold) & (varVals < var.upperThreshold) & (classVals == classOfInterest), outFeats, 0)
        
    rat.writeColumn(ratDataset, outFeatsCol, outFeats)
    ratDataset = None


def _ratapplier_defClassNames(info, inputs, outputs, otherargs):
    classNum = getattr(inputs.inrat, otherargs.classNumCol)
    
    classNames = numpy.empty_like(classNum, dtype=numpy.dtype('a255'))
    classNames[...] = ''
    
    for key in otherargs.classNamesDict:
        classNames = numpy.where((classNum == key), otherargs.classNamesDict[key], classNames)

    setattr(outputs.outrat, otherargs.classNameCol, classNames)


def defineClassNames(clumps, classNumCol, classNameCol, classNamesDict):
    """
A function to create a class names column in a RAT based on segmented clumps where a number of clumps
have the same number class.

:param clumps: input clumps image.
:param classNumCol: column specifying the class number (e.g., where clumps are segments in a segmentation)
:param classNameCol: the output column name where a string will be created if it doesn't already exists.
:param classNamesDict: Dictionary to look up the class names. The key needs to the integer number for the class

"""
    in_rats = ratapplier.RatAssociations()
    out_rats = ratapplier.RatAssociations()
                
    in_rats.inrat = ratapplier.RatHandle(clumps)
    out_rats.outrat = ratapplier.RatHandle(clumps)

    otherargs = ratapplier.OtherArguments()
    otherargs.classNumCol = classNumCol
    otherargs.classNameCol = classNameCol
    otherargs.classNamesDict = classNamesDict

    ratapplier.apply(_ratapplier_defClassNames, in_rats, out_rats, otherargs)


def setClassNamesColours(clumpsImg, classNamesCol, classInfoDict):
    """
A function to define a class names column and define the class colours.

classInfoDict = dict()
classInfoDict[1] = {'classname':'Forest', 'red':0, 'green':255, 'blue':0}
classInfoDict[2] = {'classname':'Water', 'red':0, 'green':0, 'blue':255}

:param clumpsImg: Input clumps image - expecting a classification (rather than segmentation)
                  where the number is the pixel value.
:param classNamesCol: The output column for the class names.
:param classInfoDict: a dict where the key is the pixel value for the class.

"""
    # Check numpy is available
    if not haveNumpy:
        raise Exception("The numpy module is required for this function could not be imported\n\t" + numErr)
    # Check gdal is available
    if not haveGDALPy:
        raise Exception("The GDAL python bindings are required for this function could not be imported\n\t" + gdalErr)
    # Check rios rat is available
    if not haveRIOSRat:
        raise Exception("The RIOS rat tools are required for this function could not be imported\n\t" + riosRatErr)

    n_rows = rsgislib.rastergis.getRATLength(clumpsImg)
    col_names = rsgislib.rastergis.getRATColumns(clumpsImg)

    red_avail = False
    green_avail = False
    blue_avail = False
    if 'Red' in col_names:
        red_avail = True
    if 'Green' in col_names:
        green_avail = True
    if 'Blue' in col_names:
        blue_avail = True

    class_names_col_avail = False
    if classNamesCol in col_names:
        class_names_col_avail = True

    ratDataset = gdal.Open(clumpsImg, gdal.GA_Update)

    if red_avail:
        red_arr = rat.readColumn(ratDataset, 'Red')
    else:
        red_arr = numpy.zeros(n_rows, dtype=int)

    if green_avail:
        green_arr = rat.readColumn(ratDataset, 'Green')
    else:
        green_arr = numpy.zeros(n_rows, dtype=int)

    if blue_avail:
        blue_arr = rat.readColumn(ratDataset, 'Blue')
    else:
        blue_arr = numpy.zeros(n_rows, dtype=int)

    if class_names_col_avail:
        class_names_arr = rat.readColumn(ratDataset, classNamesCol)
    else:
        class_names_arr = numpy.zeros(n_rows, dtype=numpy.dtype('a255'))

    for class_key in classInfoDict:
        if (class_key >= 0) and (class_key < n_rows):
            class_names_arr[class_key] = classInfoDict[class_key]['classname']
            red_arr[class_key] = classInfoDict[class_key]['red']
            green_arr[class_key] = classInfoDict[class_key]['green']
            blue_arr[class_key] = classInfoDict[class_key]['blue']
        else:
            print("Class key ({}) was not within the number of rows in the RAT.".format(class_key), file=sys.stderr)

    rat.writeColumn(ratDataset, classNamesCol, class_names_arr)
    rat.writeColumn(ratDataset, 'Red', red_arr)
    rat.writeColumn(ratDataset, 'Green', green_arr)
    rat.writeColumn(ratDataset, 'Blue', blue_arr)

    ratDataset = None


def populateClumpsWithClassTraining(clumpsImg, classesDict, tmpPath, classesIntCol, classesNameCol):
    """
A function to populate a clumps file with training from a series of shapefiles (1 per class)

Where:

:param clumpsImg: input clumps file.
:param classesDict: A dict structure with the class names as keys and the values are an array of two values
                   [int class val, file path for shapefile].
:param tmpPath: File path (which needs to exist) where files can temporally be written.
:param classesIntCol: Output column name for integer values representing each class.
:param classesNameCol: Output column name for string class names.

    """
    createdDIR = False
    if not os.path.isdir(tmpPath):
        os.makedirs(tmpPath)
        createdDIR = True
    
    rsPyUtils = rsgislib.RSGISPyUtils()
    uid = rsPyUtils.uidGenerator(10)
    
    classLayerSeq = list()
    tmpClassImgLayers = list()
    classNamesDict = dict()
    
    for key in classesDict:
        className = key
        classShpFile = classesDict[key][1]
        classShpFileLyr = os.path.splitext(os.path.basename(classShpFile))[0]
        classImgFile = os.path.join(tmpPath, className+"_"+uid+".kea")
        classIntVal = classesDict[key][0]
        vectorutils.rasteriseVecLyr(classShpFile, classShpFileLyr, clumpsImg, classImgFile, gdalformat="KEA",
                                    burnVal=classIntVal, datatype=rsgislib.TYPE_16UINT)
        tmpClassImgLayers.append(classImgFile)
        classNamesDict[classIntVal] = className
    
    combinedClassesImage = os.path.join(tmpPath, "CombinedClasses_" + uid + ".kea")
    imageutils.combineImages2Band(tmpClassImgLayers, combinedClassesImage, 'KEA', rsgislib.TYPE_8UINT, 0.0)
    
    rastergis.populateRATWithMode(valsimage=combinedClassesImage, clumps=clumpsImg, outcolsname=classesIntCol,
                                  usenodata=False, nodataval=0, outnodata=False, modeband=1, ratband=1)
    defineClassNames(clumpsImg, classesIntCol, classesNameCol, classNamesDict)
    
    for file in tmpClassImgLayers:
        rsPyUtils.deleteFileWithBasename(file)
    rsPyUtils.deleteFileWithBasename(combinedClassesImage)
    
    if createdDIR:
        shutil.rmtree(tmpPath)
    

def createClumpsSHPBBOX(clumpsImg, minXCol, maxXCol, minYCol, maxYCol, outShpLyrName, roundInt=False, ignoreFirstRow=False):
    """
A function to create a shapefile of polygons with the bboxs of the clumps defined using 
the minX, maxX, minY and maxY coordinates for the features.

Where:

:param clumpsImg: input clumps file.
:param minXCol: the minX column in RAT.
:param maxXCol: the maxX column in RAT.
:param minYCol: the minY column in RAT.
:param maxYCol: the maxY column in RAT.
:param outShpLyrName: The output shapefile name (layer name do not include the .shp it will be appended).
:param roundInt: Boolean specifying whether the coordinated should be rounded to integers (Default: False)

    """   
    
    # Check numpy is available
    if not haveNumpy:
        raise Exception("The numpy module is required for this function could not be imported\n\t" + numErr)
    # Check gdal is available
    if not haveGDALPy:
        raise Exception("The GDAL python bindings are required for this function could not be imported\n\t" + gdalErr)
    # Check rios rat is available
    if not haveRIOSRat:
        raise Exception("The RIOS rat tools are required for this function could not be imported\n\t" + riosRatErr)
    # Check gdal ogr is available
    if not haveGDALOGRPy:
        raise Exception("The GDAL OGR python bindings are required for this function could not be imported\n\t" + gdalogrErr)
    
    ratDataset = gdal.Open(clumpsImg)

    minXVals = rat.readColumn(ratDataset, minXCol)
    maxXVals = rat.readColumn(ratDataset, maxXCol)
    minYVals = rat.readColumn(ratDataset, minYCol)
    maxYVals = rat.readColumn(ratDataset, maxYCol)
    
    fidVals = numpy.arange(maxYVals.shape[0])

    ## Remove First Row which is no data...
    if ignoreFirstRow:    
        minXVals = numpy.delete(minXVals, (0), axis=0)
        maxXVals = numpy.delete(maxXVals, (0), axis=0)
        minYVals = numpy.delete(minYVals, (0), axis=0)
        maxYVals = numpy.delete(maxYVals, (0), axis=0)
        fidVals = numpy.delete(fidVals, (0), axis=0)
    
    ## Remove any features which are all zero (i.e., polygon not present...
    minXValsSub = minXVals[numpy.logical_not((minXVals == 0) & (maxXVals == 0) & (minYVals == 0) & (maxYVals == 0))]
    maxXValsSub = maxXVals[numpy.logical_not((minXVals == 0) & (maxXVals == 0) & (minYVals == 0) & (maxYVals == 0))]
    minYValsSub = minYVals[numpy.logical_not((minXVals == 0) & (maxXVals == 0) & (minYVals == 0) & (maxYVals == 0))]
    maxYValsSub = maxYVals[numpy.logical_not((minXVals == 0) & (maxXVals == 0) & (minYVals == 0) & (maxYVals == 0))]
    fidValsSub = fidVals[numpy.logical_not((minXVals == 0) & (maxXVals == 0) & (minYVals == 0) & (maxYVals == 0))]
        
    if roundInt:
        minXValsSub = numpy.rint(minXValsSub)
        maxXValsSub = numpy.rint(maxXValsSub)
        minYValsSub = numpy.rint(minYValsSub)
        maxYValsSub = numpy.rint(maxYValsSub)
    
    numFeats = minXValsSub.shape[0]
    print("Num Feats: ", numFeats)
    
    driver = ogr.GetDriverByName("ESRI Shapefile")
    if os.path.exists(outShpLyrName+".shp"):
        driver.DeleteDataSource(outShpLyrName+".shp")
    outDatasource = driver.CreateDataSource(outShpLyrName+ ".shp")
    raster_srs = osr.SpatialReference()
    raster_srs.ImportFromWkt(ratDataset.GetProjectionRef())
    outLayer = outDatasource.CreateLayer(outShpLyrName, srs=raster_srs)
    
    fieldFIDDefn = ogr.FieldDefn('ID', ogr.OFTInteger) 
    fieldFIDDefn.SetWidth(6) 
    outLayer.CreateField(fieldFIDDefn)
    
    print("Create and Add Polygons...")
    for i in range(numFeats):
        wktStr = "POLYGON((" + str(minXValsSub[i]) + " " + str(maxYValsSub[i]) + ", " + str(maxXValsSub[i]) + " " + str(maxYValsSub[i]) + ", " + str(maxXValsSub[i]) + " " + str(minYValsSub[i]) + ", " + str(minXValsSub[i]) + " " + str(minYValsSub[i]) + ", " + str(minXValsSub[i]) + " " + str(maxYValsSub[i]) + "))"
        #print(str(i) + ": " + wktStr)
        poly = ogr.CreateGeometryFromWkt(wktStr)
        feat = ogr.Feature( outLayer.GetLayerDefn())
        feat.SetField("ID", float(fidValsSub[i]))
        feat.SetGeometry(poly)
        if outLayer.CreateFeature(feat) != 0:
            print(str(i) + ": " + wktStr)
            print("Failed to create feature in shapefile.\n")
            sys.exit( 1 )
        feat.Destroy()
        
    outDatasource.Destroy()
    ratDataset = None
    print("Completed")


def identifySmallUnits(clumpsImg, classCol, tmpPath, outColName, smallClumpsThres, useTiledClump=False, nCores=2, tileWidth=4000, tileHeight=4000):
    """
Identify small connected units within a classification. The threshold to define small
is provided by the user in pixels. Note, the outColName and smallClumpsThres variables
can be provided as lists to identify a number of thresholds of small units.

:param clumpsImg: string for the clumps image file containing input classification
:param classCol: string for the column name representing the classification as integer values
:param tmpPath: directory path where temporary layers are stored (if directory is created within the function it will be deleted once function is complete).
:param outColName: a list of output column names (i.e., one for each threshold)
:param smallClumpsThres: a list of thresholds for identifying small clumps.
:param useTiledClump: a boolean to specify whether the tiled clumping algorithm should be used (Default is False; select True for large datasets)
:param nCores: if the tiled version of the clumping algorithm is being used then there is an option to use multiple processing cores; specify the number to be used (Default is 2).
:param tileWidth: is the width of the image tile (in pixels) if tiled clumping is used.
:param tileHeight: is the height of the image tile (in pixels) if tiled clumping is used.

Example::

    from rsgislib.rastergis import ratutils
    
    clumpsImg = "LS2MSS_19750620_lat10lon6493_r67p250_rad_srefdem_30m_clumps.kea"
    tmpPath = "./tmp/"
    classCol = "OutClass"
    outColName = ["SmallUnits25", "SmallUnits50", "SmallUnits100"]
    smallClumpsThres = [25, 50, 100]
    rastergis.identifySmallUnits(clumpsImg, classCol, tmpPath, outColName, smallClumpsThres)

    """
    
    # Check numpy is available
    if not haveNumpy:
        raise Exception("The numpy module is required for this function could not be imported\n\t" + numErr)
    # Check gdal is available
    if not haveGDALPy:
        raise Exception("The GDAL python bindings are required for this function could not be imported\n\t" + gdalErr)
    # Check rios rat is available
    if not haveRIOSRat:
        raise Exception("The RIOS rat tools are required for this function could not be imported\n\t" + riosRatErr)
    
    if len(outColName) is not len(smallClumpsThres):
        print("The number of threshold values and output column names should be the same.")
        sys.exit(-1)
    
    numThresholds = len(smallClumpsThres)
        
    createdDIR = False
    if not os.path.isdir(tmpPath):
        os.makedirs(tmpPath)
        createdDIR = True
    
    baseName = os.path.splitext(os.path.basename(clumpsImg))[0]
    classMaskImg = os.path.join(tmpPath, baseName+"_TmpClassMask.kea")
    classMaskClumps = os.path.join(tmpPath, baseName+"_TmpClassMaskClumps.kea")
    smallClumpsMask = os.path.join(tmpPath, baseName+"_SmallClassClumps.kea")
    
    rastergis.exportCol2GDALImage(clumpsImg, classMaskImg, "KEA", rsgislib.TYPE_16UINT, classCol)
    if useTiledClump:
        from rsgislib.segmentation import tiledclump
        if nCores > 1:
            tiledclump.performClumpingMultiProcess(classMaskImg, classMaskClumps, tmpDIR=os.path.join(tmpPath, baseName+'_ClumpTmp'), width=tileWidth, height=tileHeight, nCores=nCores)
        else:
            tiledclump.performClumpingSingleThread(classMaskImg, classMaskClumps, tmpDIR=os.path.join(tmpPath, baseName+'_ClumpTmp', width=tileWidth, height=tileHeight))
    else:
        segmentation.clump(classMaskImg, classMaskClumps, "KEA", False, 0)
    rastergis.populateStats(classMaskClumps, False, False)
        
    for i in range(numThresholds):
        print("Processing thresold " + str(smallClumpsThres[i]) + " - " + outColName[i])
        ratDataset = gdal.Open(classMaskClumps, gdal.GA_Update)
        Histogram = rat.readColumn(ratDataset, "Histogram")
        smallUnits = numpy.zeros_like(Histogram, dtype=numpy.int16)
        smallUnits[Histogram < smallClumpsThres[i]] = 1
        rat.writeColumn(ratDataset, "smallUnits", smallUnits)
        ratDataset = None
    
        rastergis.exportCol2GDALImage(classMaskClumps, smallClumpsMask, "KEA", rsgislib.TYPE_8UINT, "smallUnits")
    
        bs = []
        bs.append(rastergis.BandAttStats(band=1, maxField=outColName[i]))
        rastergis.populateRATWithStats(smallClumpsMask, clumpsImg, bs)
    
    rsgisUtils = rsgislib.RSGISPyUtils()
    rsgisUtils.deleteFileWithBasename(classMaskImg)
    rsgisUtils.deleteFileWithBasename(classMaskClumps)
    rsgisUtils.deleteFileWithBasename(smallClumpsMask)
    if createdDIR:
        shutil.rmtree(tmpPath)



def _computeProximityArrArgsFunc(argVals):
    """
    This function is used internally within calcDist2Classes for the multiprocessing Pool

    """
    classImgDS = gdal.Open(argVals[0], gdal.GA_ReadOnly)
    classImgBand = classImgDS.GetRasterBand(1)
    imageutils.createCopyImage(argVals[0], argVals[1], 1, argVals[3], argVals[4], rsgislib.TYPE_32FLOAT)
    distImgDS = gdal.Open(argVals[1], gdal.GA_Update)
    distImgBand = distImgDS.GetRasterBand(1)
    gdal.ComputeProximity(classImgBand, distImgBand, argVals[2], callback=gdal.TermProgress)
    distImgBand = None
    distImgDS = None
    classImgBand = None
    classImgDS = None



def calcDist2Classes(clumpsImg, classCol, outImgBase, tmpDIR='./tmp', tileSize=2000, maxDist=1000, nodata=1000, nCores=-1):
    """
    A function which will calculate proximity rasters for a set of classes defined within the RAT.
    
    :param clumpsImg: is a string specifying the input image with the associated RAT
    :param classCol: is the column in the RAT which has the classification
    :param outImgBase: is the base name of the output image - output files will be KEA files.
    :param tmpDIR: is a directory to be used for storing the image tiles and other temporary files - if not directory does not exist it will be created and deleted on completion (Default: ./tmp).
    :param tileSize: is an int specifying in pixels the size of the image tiles used for processing (Default: 2000)
    :param maxDist: is the maximum distance in units of the geographic units of the projection of the input image (Default: 1000).
    :param nodata: is the value applied to the pixels outside of the maxDist threshold (Default: 1000; i.e., the same as maxDist).
    :param nCores: is the number of processing cores which are available to be used for this processing. If -1 all available cores will be used. (Default: -1)

    """
    tmpPresent = True
    if not os.path.exists(tmpDIR):
        print("WARNING: \'" + tmpDIR + "\' directory does not exist so creating it...")
        os.makedirs(tmpDIR)
        tmpPresent = False
    
    if nCores <= 0:
        nCores = multiprocessing.cpu_count()
    
    rsgisUtils = rsgislib.RSGISPyUtils()
    uid = rsgisUtils.uidGenerator()
        
    classesImg = os.path.join(tmpDIR, 'ClassImg_'+uid+'.kea')
    rastergis.exportCol2GDALImage(clumpsImg, classesImg, 'KEA', rsgislib.TYPE_32UINT, classCol)
    
    ratDataset = gdal.Open(clumpsImg, gdal.GA_ReadOnly)
    classColInt = rat.readColumn(ratDataset, classCol)
    ratDataset = None

    classIDs = numpy.unique(classColInt)
    
    xRes, yRes = rsgisUtils.getImageRes(classesImg)
    
    #print("Image Res {} x {}".format(xRes, yRes))
    
    xMaxDistPxl = math.ceil(maxDist/xRes)
    yMaxDistPxl = math.ceil(maxDist/yRes)
    
    print("Max Dist Pxls X = {}, Y = {}".format(xMaxDistPxl, yMaxDistPxl))
    
    tileOverlap = xMaxDistPxl
    if yMaxDistPxl > xMaxDistPxl:
        tileOverlap = yMaxDistPxl
    
    classTilesDIR = os.path.join(tmpDIR, 'ClassTiles_'+uid)
    classTilesDIRPresent = True
    if not os.path.exists(classTilesDIR):
        os.makedirs(classTilesDIR)
        classTilesDIRPresent = False
    
    classesImgTileBase = os.path.join(classTilesDIR, 'ClassImgTile')
    imageutils.createTiles(classesImg, classesImgTileBase, tileSize, tileSize, tileOverlap, 0, 'KEA', rsgislib.TYPE_32UINT, 'kea')
    imgTileFiles = glob.glob(classesImgTileBase+'*.kea')
    
    distTilesDIR = os.path.join(tmpDIR, 'DistTiles_'+uid)
    distTilesDIRPresent = True
    if not os.path.exists(distTilesDIR):
        os.makedirs(distTilesDIR)
        distTilesDIRPresent = False
    
    proxOptionsBase = ['MAXDIST='+str(maxDist), 'DISTUNITS=GEO', 'NODATA='+str(nodata)]
    
    for classID in classIDs:
        print("Class {}".format(classID))
        proxOptions = list(proxOptionsBase)
        proxOptions.append('VALUES='+str(classID))

        distTiles = []
        distTileArgs = []
        for classTileFile in imgTileFiles:
            baseTileName = os.path.basename(classTileFile)
            distTileFile = os.path.join(distTilesDIR, baseTileName)
            tileArgs = [classTileFile, distTileFile, proxOptions, nodata, 'KEA']
            distTiles.append(distTileFile)
            distTileArgs.append(tileArgs)
        
        with Pool(nCores) as p:
            p.map(_computeProximityArrArgsFunc, distTileArgs)
                
        distImage = outImgBase + '_' + str(classID) + '.kea'
        # Mosaic Tiles
        imageutils.createImageMosaic(distTiles, distImage, nodata, nodata, 1, 1, 'KEA', rsgislib.TYPE_32FLOAT)
        imageutils.popImageStats(distImage, usenodataval=True, nodataval=nodata, calcpyramids=True)
        for imgFile in distTiles:
            rsgisUtils.deleteFileWithBasename(imgFile)
    
    if not classTilesDIRPresent:
        shutil.rmtree(classTilesDIR, ignore_errors=True)
    else:
        for classTileFile in imgTileFiles:
            rsgisUtils.deleteFileWithBasename(classTileFile)
    
    if not distTilesDIRPresent:
        shutil.rmtree(distTilesDIR, ignore_errors=True)
    
    if not tmpPresent:
        shutil.rmtree(tmpDIR, ignore_errors=True)
    else:
        os.remove(classesImg)


def calcDistBetweenClumps(clumpsImg, outColName, tmpDIR='./tmp', useIdx=False, maxDistThres=10):
    """
Calculate the distance between all clumps

:param clumpsImg: image clumps for which the distance will be calculated.
:param outColName: output column within the clumps image.
:param tmpDIR: directory out tempory files will be outputted to.
:param useIdx: use a spatial index when calculating the distance between clumps (needed for large number of clumps).
:param maxDistThres: if using an index than an upper limit on the distance between clumps can be defined.

"""
    tmpPresent = True
    if not os.path.exists(tmpDIR):
        os.makedirs(tmpDIR)
        tmpPresent = False 
    
    rsgisUtils = rsgislib.RSGISPyUtils()
    
    baseName = os.path.splitext(os.path.basename(clumpsImg))[0]+'_'+rsgisUtils.uidGenerator()
    
    polysShp = os.path.join(tmpDIR, baseName+'_shp.shp')
    vectorutils.polygoniseRaster(clumpsImg, polysShp, imgBandNo=1, maskImg=clumpsImg, imgMaskBandNo=1)
    
    print('Calculating Distance - can take some time. Try using index and decreasing max distance threshold.')
    t = rsgislib.RSGISTime()
    t.start(True)
    polysShpGeomDist = os.path.join(tmpDIR, baseName+'_dist_shp.shp')
    vectorutils.dist2NearestGeom(polysShp, polysShpGeomDist, outColName, True, useIdx, maxDistThres)
    t.end()
    
    rastergis.importVecAtts(clumpsImg, polysShpGeomDist, [outColName])
    
    if not tmpPresent:
        shutil.rmtree(tmpDIR, ignore_errors=True)


def calcDistToLargeClumps(clumpsImg, outColName, sizeThres, tmpDIR='./tmp', useIdx=False, maxDistThres=10):
    """
Calculate the distance from each small clump to a large clump. Split defined by the threshold provided.

:param clumpsImg: image clumps for which the distance will be calculated.
:param outColName: output column within the clumps image.
:param sizeThres: is a threshold to seperate the sets of large and small clumps.
:param tmpDIR: directory out tempory files will be outputted to.
:param useIdx: use a spatial index when calculating the distance between clumps (needed for large number of clumps).
:param maxDistThres: if using an index than an upper limit on the distance between clumps can be defined.

"""
    tmpPresent = True
    if not os.path.exists(tmpDIR):
        os.makedirs(tmpDIR)
        tmpPresent = False 
    
    rsgisUtils = rsgislib.RSGISPyUtils()
    uidStr = rsgisUtils.uidGenerator()
    
    baseName = os.path.splitext(os.path.basename(clumpsImg))[0]+'_'+uidStr
    
    ratDataset = gdal.Open(clumpsImg, gdal.GA_Update)
    Histogram = rat.readColumn(ratDataset, "Histogram")
    smallUnits = numpy.zeros_like(Histogram, dtype=numpy.int16)
    smallUnits[Histogram < sizeThres] = 1
    rat.writeColumn(ratDataset, "smallUnits", smallUnits)
    
    ID = numpy.arange(Histogram.shape[0])
    
    smUnitIDs = ID[smallUnits == 1]
    smUnitIDs = smUnitIDs[smUnitIDs>0]    
    lrgUnitIDs = ID[smallUnits == 0]
    lrgUnitIDs = lrgUnitIDs[lrgUnitIDs>0]
    
    print("There are {} small clumps.".format(smUnitIDs.shape[0]))
    print("There are {} large clumps.".format(lrgUnitIDs.shape[0]))
    
    smUnitFIDs = numpy.arange(smUnitIDs.shape[0])
    lrgUnitFIDs = numpy.arange(lrgUnitIDs.shape[0])
    
    smUnitClumpIDs = numpy.zeros_like(Histogram, dtype=numpy.int16)
    smUnitClumpIDs[smUnitIDs] = smUnitFIDs
    lrgUnitClumpIDs = numpy.zeros_like(Histogram, dtype=numpy.int16)
    lrgUnitClumpIDs[lrgUnitIDs] = lrgUnitFIDs
    
    rat.writeColumn(ratDataset, "SmUnits", smUnitClumpIDs)
    rat.writeColumn(ratDataset, "LrgUnits", lrgUnitClumpIDs)
    rat.writeColumn(ratDataset, "smallUnitsBin", smallUnits)    
    
    smClumpsImg = os.path.join(tmpDIR, baseName+'_smclumps.kea')
    rastergis.exportCol2GDALImage(clumpsImg, smClumpsImg, 'KEA', rsgislib.TYPE_32UINT, 'SmUnits')
    rastergis.populateStats(clumps=smClumpsImg, addclrtab=True, calcpyramids=True, ignorezero=True)
    
    lrgClumpsImg = os.path.join(tmpDIR, baseName+'_lrgclumps.kea')
    rastergis.exportCol2GDALImage(clumpsImg, lrgClumpsImg, 'KEA', rsgislib.TYPE_32UINT, 'LrgUnits')
    rastergis.populateStats(clumps=lrgClumpsImg, addclrtab=True, calcpyramids=True, ignorezero=True)
    
    smPolysShp = os.path.join(tmpDIR, baseName+'_smClumps_shp.shp')
    rsgislib.vectorutils.polygoniseRaster(smClumpsImg, smPolysShp, imgBandNo=1, maskImg=smClumpsImg, imgMaskBandNo=1)
    
    lgrPolysShp = os.path.join(tmpDIR, baseName+'_lgrClumps_shp.shp')
    rsgislib.vectorutils.polygoniseRaster(lrgClumpsImg, lgrPolysShp, imgBandNo=1, maskImg=lrgClumpsImg, imgMaskBandNo=1)
    
    print('Calculating Distance - can take some time. Try using index and decreasing max distance threshold.')
    t = rsgislib.RSGISTime()
    t.start(True)
    smPolysDistShp = os.path.join(tmpDIR, baseName+'_smClumps_dist_shp.shp')
    rsgislib.vectorutils.dist2NearestSecGeomSet(smPolysShp, lgrPolysShp, smPolysDistShp, outColName, True, useIdx, maxDistThres)
    t.end()
    rsgislib.rastergis.importVecAtts(smClumpsImg, smPolysDistShp, [outColName])
    
    smClumpsRATDataset = gdal.Open(smClumpsImg, gdal.GA_Update)
    minDistCol = rat.readColumn(smClumpsRATDataset, outColName)
        
    minDistSmlClumpsArr = numpy.zeros_like(Histogram, dtype=numpy.float32)
    minDistSmlClumpsArr[smUnitIDs] = minDistCol
    
    rat.writeColumn(ratDataset, outColName, minDistSmlClumpsArr)
    
    smClumpsRATDataset = None
    ratDataset = None
    
    if not tmpPresent:
        shutil.rmtree(tmpDIR, ignore_errors=True)


def takeRandomSample(clumpsImg, inColName, inColVal, outColName, sampleRatio, seed=0):
    """
A function to take a random sample of an input column.

:param clumpsImg: clumps image.
:param inColName: input column name.
:param inColVal: numeric value for which the random sample is to be taken for.
:param outColName: output column where value of 1 is selected within the random sample and 0 is not selected.
:param sampleRatio: the size of the sample (0 - 1.0; i.e., 10% = 0.1) to be taken of the number of rows within input value.
:param seed: is the seed for the random number generation (optional; default is 0).

"""
    if (sampleRatio <= 0) or (sampleRatio >= 1):
        raise rsgislib.RSGISPyException("Input sample ratio must be between 0 and 1.")
    
    # Define the seed for the random number generation.
    numpy.random.seed(seed)
    
    # READ COL FROM RAT:
    ratDataset = gdal.Open(clumpsImg, gdal.GA_Update)
    inColVals = rat.readColumn(ratDataset, inColName)
    
    # Create an array for the original array indexes.
    ID = numpy.arange(inColVals.shape[0])
    
    # Cut array and array indexes to be only bins with a value of 1 (i.e., there is lidar data here).
    ID = ID[inColVals==inColVal]
    
    # Sample of the subsetted input array where it equals inColVal.
    numOfSamples = int(ID.shape[0]*sampleRatio)
    IDSamples = numpy.arange(ID.shape[0])
    IDSampPerms = numpy.random.permutation(IDSamples)
    IDSampPermsSelect = IDSampPerms[0:numOfSamples]
    
    # Find the array indexes for the whole input array (i.e., the whole RAT).
    outArryIdxsSel = ID[IDSampPermsSelect]
    
    # Create output columns for writing to RAT
    outColVals = numpy.zeros_like(inColVals)
    
    # Populate columns where those selected have value 1.
    outColVals[outArryIdxsSel] = 1
    
    # WRITE COL TO RAT:
    rat.writeColumn(ratDataset, outColName, outColVals)
    ratDataset = None


def getColumnData(clumpsImg, columnName):
    """
A function to read a column of data from a RAT.

:param clumpsImg: Input clumps image
:param columnName: Name of the column to be read.

:return: numpy array with values from the clumpsImg

"""
    # Check numpy is available
    if not haveNumpy:
        raise Exception("The numpy module is required for this function could not be imported\n\t" + numErr)
    # Check gdal is available
    if not haveGDALPy:
        raise Exception("The GDAL python bindings are required for this function could not be imported\n\t" + gdalErr)
    # Check rios rat is available
    if not haveRIOSRat:
        raise Exception("The RIOS rat tools are required for this function could not be imported\n\t" + riosRatErr)

    col_names = rsgislib.rastergis.getRATColumns(clumpsImg)

    if columnName not in col_names:
        raise Exception("Column specified is not within the RAT.")

    ratDataset = gdal.Open(clumpsImg, gdal.GA_ReadOnly)
    if ratDataset is None:
        raise Exception("The input image could not be opened.")

    col_data = rat.readColumn(ratDataset, columnName)
    ratDataset = None
    return col_data


def setColumnData(clumpsImg, columnName, columnData):
    """
A function to read a column of data from a RAT.

:param clumpsImg: Input clumps image
:param columnName: Name of the column to be written.
:param columnData: Data to be written to the column.

"""
    # Check numpy is available
    if not haveNumpy:
        raise Exception("The numpy module is required for this function could not be imported\n\t" + numErr)
    # Check gdal is available
    if not haveGDALPy:
        raise Exception("The GDAL python bindings are required for this function could not be imported\n\t" + gdalErr)
    # Check rios rat is available
    if not haveRIOSRat:
        raise Exception("The RIOS rat tools are required for this function could not be imported\n\t" + riosRatErr)

    rat_length = rsgislib.rastergis.getRATLength(clumpsImg)
    if rat_length != (columnData.shape[0]):
        raise Exception("The input data array is not the same length as the RAT.")

    ratDataset = gdal.Open(clumpsImg, gdal.GA_Update)
    if ratDataset is None:
        raise Exception("The input image could not be opened.")

    rat.writeColumn(ratDataset, columnName, columnData)
    ratDataset = None

